# pylint: disable=abstract-method
"""
{{ cookiecutter.description }}
"""
import math
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests_cache

from shillelagh.adapters.base import Adapter
from shillelagh.fields import Field
from shillelagh.filters import Filter, Operator

# from shillelagh.lib import SimpleCostModel
from shillelagh.typing import RequestedOrder, Row

AVERAGE_NUMBER_OF_ROWS = 1000
FIXED_COST = 0


class {{ cookiecutter.adapter_name|replace(' ', '') }}{{ cookiecutter.adapter_type }}(Adapter):

    """
    {{ cookiecutter.description }}
    """

    # Set this to ``True`` if the adapter doesn't access the filesystem.
    safe = False

    # If this is true, the adapter will receive a ``limit`` argument in the
    # ``get_data`` method, and will be responsible for limiting the number of
    # rows returned.
    supports_limit = False

    # Similarly, if this is true the adapter will receive an ``offset`` argument
    # in the ``get_data`` method, and will be responsible for offsetting the
    # rows that are returned.
    supports_offset = False

    # If this is true, the adapter will receive a ``requested_columns`` argument
    # in the ``get_data`` method, and will be responsible for returning only the
    # requested columns. Otherwise the adapter will have no indication of which
    # columns were requested, and should return all columns always.
    supports_requested_columns = False

    # This method is used to determine which URIs your adapter will handle. For
    # example, if your adapter interfaces with an API at api.example.com you
    # can do:
    #
    #     parsed = urllib.parse.urlparse(uri)
    #     return parsed.netloc == "api.example.com"
    #
    # The method will be called 1 or 2 times. On the first call ``fast`` will be
    # true, and the adapter should return as soon as possible. This means that if
    # you need to do a network request to determine if the URI should be handled
    # by your adapter, when ``fast=True`` you should return ``None``.
    #
    # On the second call, ``fast`` will be false, and the adapter can make network
    # requests to introspect URIs. The adapter should then return either true or
    # false on the second request.
    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        return False

    # This method parses the URI into arguments that are passed to initialize the
    # adapter class. The simplest implementation returns the URI unmodified, but
    # adapters might process the URL and return only relevant parts of the URI,
    # like an ID or path.
    @staticmethod
    def parse_uri(uri: str) -> Tuple[str]:
        return (uri,)

    def __init__(self, uri: str):
        super().__init__()
        self.uri = uri

        # If the adapter needs to do API requests it's useful to use a cache for
        # the requests. If you're not doing network requests you get delete this
        # session object.
        self._session = requests_cache.CachedSession(
            cache_name="{{ cookiecutter.slug }}_cache",
            backend="sqlite",
            expore_after=180,
        )

        # For adapters with dynamic columns (ie, number, names, and types of
        # columns depend on the URI) it's good practice to set the columns when
        # the class is initialized.
        self._set_columns()

    # When defining columns it's important to know which columns can be used to
    # filter the URI. For example, if the API accepts a time range the adapter
    # probably needs a temporal column that is filterable:
    #
    #     from shillelagh.fields import DateTime
    #     from shillelagh.filters import Range
    #
    #     self.columns["time"] = DateTime(filters=[Range])
    #
    # It's then up to the ``get_data`` method to translate temporal filters into
    # the corresponding API calls.
    #
    # The column definition should also specify if the column has a natural order,
    # or if the adapter can handle any requested order (and the adapter will be
    # responsible for ordering the data):
    #
    #     from shillelagh.fields import Integer
    #     from shillelagh.fields import Order
    #
    #     self.columns["prime_numbers"] = Integer(
    #         filters=[Range],
    #         order=Order.ASCENDING,
    #     )
    #
    # Finally, columns can return data that is not filtered perfectly — eg, hourly
    # data filtered only down to the daily granularity. In that case the column
    # should be declared as inexact:
    #
    #     self.columns["time"] = DateTime(filters=[Range], exact=False)
    #
    def _set_columns(self) -> None:
        self.columns: Dict[str, Field] = {}

    # If you have static columns you can get rid of the ``get_columns`` method,
    # get rid of ``_set_columns``, and simply define the columns as class
    # attributes:
    #
    # time = DateTime(filters=[Range])
    #
    # Then ``get_columns`` from the parent class will find these attributes and
    # use them as the dictionary of columns.
    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    # This method is used to return any extra metadata associated with the URI.
    # You can delete it instead of returning an empty dictionary, since that's
    # the exact implementation of the parent method.
    def get_metadata(self) -> Dict[str, Any]:
        return {}

    # A method for estimating the cost of a query, used by the query planner. The
    # model receives the name of the columns and the operators filtering them, as
    # well as the columns that should be sorted and how.
    def get_cost(
        self,
        filtered_columns: List[Tuple[str, Operator]],
        order: List[Tuple[str, RequestedOrder]],
    ) -> float:
        # A simple model for estimating query costs.
        #
        # The model assumes that each filtering operation is O(n), and each
        # sorting operation is O(n log n), in addition to a fixed cost.
        return int(
            FIXED_COST
            + AVERAGE_NUMBER_OF_ROWS * len(filtered_columns)
            + AVERAGE_NUMBER_OF_ROWS * math.log2(AVERAGE_NUMBER_OF_ROWS) * len(order)
        )

    # If you're using the cost model above unmodified you can simply point the
    # method to the ``SimpleCostModel`` class. Don't forget to uncomment the import
    # at the top of the file.
    # get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS, FIXED_COST)

    # This method yields rows of data, each row a dictionary. If any columns are
    # declared as filterable there might be a corresponding ``Filter`` object in
    # the ``bounds`` argument that must be used to filter the column (unless the
    # column was declared as inexact).
    #
    # Note that, in addition to the actual columns, each row should also have a
    # column called ``rowid``, with an integer value. If the adapter is read-only
    # this can be any number; for adapters that implement DML the rowid is used
    # in ``INSERT``, ``UPDATE``, and ``DELETE`` operations.
    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
        # if ``supports_limit`` is true uncomment below:
        # limit: Optional[int] = None,
        # if ``supports_offset` is true uncomment below:
        # offset: Optional[int] = None,
        # if ``supports_requested_columns`` is true uncomment below:
        # requested_columns: Optional[Set[str]] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        pass

    # For adapters that support ``INSERT`` statements, this method will be called
    # for every row that is inserted. Note that ``row`` will have a special column
    # called ``rowid``.
    #
    # The CSV adapter is a good example of how to handle row IDs. Each row in the
    # file is assigned a sequential row ID, starting from zero, and the CSV adapter
    # uses a ``RowIDManager`` to keep track of inserts and deletes.
    def insert_data(self, row: Row) -> int:
        pass

    # Delete a row given its ID.
    def delete_data(self, row_id: int) -> None:
        pass

    # Method for updating data. If this method is not implemented the base method
    # will perform a delete followed by an insert.
    def update_data(self, row_id: int, row: Row) -> None:
        pass

    # Close the file. This can be used for garbage collection. For example, the
    # CSV adapter uses this method to effectively delete rows marked as deleted;
    # the GSheets adapter uses this to synchronize the local copy of the sheet back
    # to the actual sheet.
    def close(self) -> None:
        pass

    # This is called when the user runs ``DROP TABLE`` on a table.
    def drop_table(self) -> None:
        pass



"""Custom types for Shillelagh."""

from typing import Any, Optional, Union

from typing_extensions import Literal

from shillelagh.fields import Field, Order

# A value corresponding to a constraint is one of:
#     None
#         This means you have no index for that constraint. SQLite will have to
#         iterate over every row for it.
#     integer
#         This is the argument number for the constraintargs being passed
#         into the Filter() function of your cursor.
#     (integer, boolean)
#         By default SQLite will check what you return. If you set the boolean
#         to False then SQLite won’t do that double checking.
Constraint = Union[None, int, tuple[int, bool]]
SQLiteConstraint = int

# A row of data
Row = dict[str, Any]

# An index is a tuple with a column index and an operator to filter it
Index = tuple[int, SQLiteConstraint]

OrderBy = tuple[int, bool]
RequestedOrder = Union[Literal[Order.ASCENDING], Literal[Order.DESCENDING]]

SQLiteValidType = Union[None, int, float, str, bytes]


# Cursor description
Description = Optional[
    list[
        tuple[
            str,
            type[Field],
            Optional[str],
            Optional[str],
            Optional[str],
            Optional[str],
            Optional[bool],
        ]
    ]
]

MaybeType = Optional[bool]
Maybe = None  # pylint: disable=invalid-name



# pylint: disable=c-extension-no-member, invalid-name
"""
A SQLite virtual table.

This module implements a SQLite virtual table that delegates data requests
to adapters. The main goal is to make the interface easier to use, to
simplify the work of writing new adapters.
"""

import json
import logging
from collections import defaultdict
from collections.abc import Iterator
from typing import Any, DefaultDict, Optional, cast

import apsw

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import (
    Blob,
    FastISODateTime,
    Field,
    Float,
    IntBoolean,
    ISODate,
    ISOTime,
    Order,
    RowID,
    String,
    StringDecimal,
    StringDuration,
    StringInteger,
)
from shillelagh.filters import Operator
from shillelagh.lib import best_index_object_available, deserialize, get_bounds
from shillelagh.typing import (
    Constraint,
    Index,
    OrderBy,
    RequestedOrder,
    Row,
    SQLiteConstraint,
    SQLiteValidType,
)

if best_index_object_available():  # pragma: no cover
    from apsw.ext import index_info_to_dict
else:  # pragma: no cover
    apsw.IndexInfo = Any  # for type annotation

    # pylint: disable=unused-argument
    def index_info_to_dict(index_info: apsw.IndexInfo) -> None:
        """
        Dummy function for testing.
        """


_logger = logging.getLogger(__name__)


# map between APSW operators and the ``Operator`` enum
operator_map = {
    apsw.SQLITE_INDEX_CONSTRAINT_EQ: Operator.EQ,
    apsw.SQLITE_INDEX_CONSTRAINT_GE: Operator.GE,
    apsw.SQLITE_INDEX_CONSTRAINT_GT: Operator.GT,
    apsw.SQLITE_INDEX_CONSTRAINT_LE: Operator.LE,
    apsw.SQLITE_INDEX_CONSTRAINT_LT: Operator.LT,
}


def _add_sqlite_constraint(constant_name: str, operator: Operator) -> None:
    if hasattr(apsw, constant_name):
        operator_map[getattr(apsw, constant_name)] = operator


# SQLITE_INDEX_CONSTRAINT_LIKE, >=3.10.0
_add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_LIKE", Operator.LIKE)
# SQLITE_INDEX_CONSTRAINT_NE, >=3.21.0
_add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_NE", Operator.NE)
# SQLITE_INDEX_CONSTRAINT_ISNULL, >=3.21.0
_add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_ISNULL", Operator.IS_NULL)
# SQLITE_INDEX_CONSTRAINT_ISNOTNULL, >=3.21.0
_add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_ISNOTNULL", Operator.IS_NOT_NULL)
# SQLITE_INDEX_CONSTRAINT_LIMIT, >=3.38.0
_add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_LIMIT", Operator.LIMIT)
# SQLITE_INDEX_CONSTRAINT_OFFSET, >=3.38.0
_add_sqlite_constraint("SQLITE_INDEX_CONSTRAINT_OFFSET", Operator.OFFSET)

# limit and offset are special constraints without an associated column index
LIMIT_OFFSET_INDEX = -1

# map for converting between Python native types (boolean, datetime, etc.)
# and types understood by SQLite (integers, strings, etc.)
type_map: dict[str, type[Field]] = {
    field.type: field  # type: ignore
    for field in [
        Blob,
        FastISODateTime,
        Float,
        ISODate,
        ISOTime,
        IntBoolean,
        String,
        StringDecimal,
        StringDuration,
        StringInteger,
    ]
}


# a row with only SQLite-valid types
SQLiteRow = dict[str, SQLiteValidType]


def convert_rows_to_sqlite(
    columns: dict[str, Field],
    rows: Iterator[Row],
) -> Iterator[SQLiteRow]:
    """
    Convert values from native Python types to SQLite types.

    Native Python types like ``datetime.datetime`` are not supported by SQLite; instead
    we need to cast them to strings or numbers. We use the original fields to handle
    the conversion (not the adapter fields).
    """
    converters = {
        column_name: type_map[column_field.type]().format
        for column_name, column_field in columns.items()
    }
    converters["rowid"] = RowID().format
    for row in rows:
        yield {
            column_name: converters[column_name](value)
            for column_name, value in row.items()
        }


def convert_rows_from_sqlite(
    columns: dict[str, Field],
    rows: Iterator[SQLiteRow],
) -> Iterator[Row]:
    """
    Convert values from SQLite types to native Python types.

    Native Python types like ``datetime.datetime`` are not supported by SQLite; instead
    we need to cast them to strings or numbers. We use the original fields to handle
    the conversion (not the adapter fields).
    """
    converters = {
        column_name: type_map[column_field.type]().parse
        for column_name, column_field in columns.items()
    }
    converters["rowid"] = RowID().parse
    for row in rows:
        yield {
            column_name: converters[column_name](value)
            for column_name, value in row.items()
        }


def get_all_bounds(
    indexes: list[Index],
    constraintargs: list[Any],
    columns: dict[str, Field],
) -> DefaultDict[str, set[tuple[Operator, Any]]]:
    """
    Convert indexes and constraints to operations on each column.
    """
    column_names = list(columns.keys())

    all_bounds: DefaultDict[str, set[tuple[Operator, Any]]] = defaultdict(set)
    for (column_index, sqlite_index_constraint), constraint in zip(
        indexes,
        constraintargs,
    ):
        if sqlite_index_constraint not in operator_map:
            # pylint: disable=broad-exception-raised
            raise Exception(f"Invalid constraint passed: {sqlite_index_constraint}")
        if column_index == LIMIT_OFFSET_INDEX:
            continue
        operator = operator_map[sqlite_index_constraint]
        column_name = column_names[column_index]
        column_type = columns[column_name]

        # convert constraint to native Python type, then to DB specific type
        constraint = type_map[column_type.type]().parse(constraint)
        value = column_type.format(constraint)

        all_bounds[column_name].add((operator, value))

    return all_bounds


def get_limit_offset(
    indexes: list[Index],
    constraintargs: list[Any],
) -> tuple[Optional[int], Optional[int]]:
    """
    Extract limit and offset.
    """
    limit = offset = None

    for (column_index, sqlite_index_constraint), constraint in zip(
        indexes,
        constraintargs,
    ):
        if sqlite_index_constraint not in operator_map:
            # pylint: disable=broad-exception-raised
            raise Exception(f"Invalid constraint passed: {sqlite_index_constraint}")
        if column_index != LIMIT_OFFSET_INDEX:
            continue
        operator = operator_map[sqlite_index_constraint]
        if operator == Operator.LIMIT:
            limit = constraint
        elif operator == Operator.OFFSET:
            offset = constraint

    return limit, offset


def get_order(
    orderbys: list[OrderBy],
    column_names: list[str],
) -> list[tuple[str, RequestedOrder]]:
    """
    Return a list of column names and sort order from a SQLite orderbys.
    """
    return [
        (
            column_names[column_index],
            Order.DESCENDING if descending else Order.ASCENDING,
        )
        for column_index, descending in orderbys
    ]


class VTModule:  # pylint: disable=too-few-public-methods
    """
    A module used to create SQLite virtual tables.

    This implementation delegates data requests to a Shillelagh adapter, simplifying
    the work needed to support new data sources.
    """

    def __init__(self, adapter: type[Adapter]):
        self.adapter = adapter

    def Create(  # pylint: disable=unused-argument
        self,
        connection: apsw.Connection,
        modulename: str,
        dbname: str,
        tablename: str,
        *args: str,
    ) -> tuple[str, "VTTable"]:
        """
        Called when a table is first created on a connection.
        """
        deserialized_args = [deserialize(arg[1:-1]) for arg in args]
        _logger.debug(
            "Instantiating adapter with deserialized arguments: %s",
            deserialized_args,
        )
        adapter = self.adapter(*deserialized_args)
        table = VTTable(adapter)
        create_table = table.get_create_table(tablename)
        return create_table, table

    Connect = Create


class VTTable:
    """
    A SQLite virtual table.

    The VTTable object contains knowledge of the indices, makes cursors and can
    perform transactions.

    A virtual table is structured as a series of rows, each of which has the same
    columns. The value in a column must be one of the 5 supported types, but the
    type can be different between rows for the same column. The virtual table
    routines identify the columns by number, starting at zero.

    Each row has a unique 64 bit integer rowid with the Cursor routines operating
    on this number, as well as some of the Table routines such as UpdateChangeRow.
    """

    def __init__(self, adapter: Adapter):
        self.adapter = adapter

    def get_create_table(self, tablename: str) -> str:
        """
        Return the table's ``CREATE TABLE`` statement.
        """
        columns = self.adapter.get_columns()
        if not columns:
            raise ProgrammingError(f"Virtual table {tablename} has no columns")

        quoted_columns = {k.replace('"', '""'): v for k, v in columns.items()}
        formatted_columns = ", ".join(
            f'"{k}" {v.type}' for (k, v) in quoted_columns.items()
        )
        return f'CREATE TABLE "{tablename}" ({formatted_columns})'

    def _build_index(  # pylint: disable=too-many-locals
        self,
        constraints: list[tuple[int, SQLiteConstraint]],
        orderbys: list[OrderBy],
    ) -> tuple[list[Constraint], int, list[Index], list[OrderBy], bool, float]:
        """
        Helper function to build index.
        """
        columns = self.adapter.get_columns()
        column_names = list(columns.keys())
        column_types = list(columns.values())

        # currently the index number is not used for anything; instead, we encode the
        # the index as JSON in ``index_name``
        index_number = 42

        indexes: list[Index] = []
        constraints_used: list[Constraint] = []
        filter_index = 0
        filtered_columns: list[tuple[str, Operator]] = []
        for column_index, sqlite_index_constraint in constraints:
            operator = operator_map.get(sqlite_index_constraint)

            # LIMIT/OFFSET
            if (operator is Operator.LIMIT and self.adapter.supports_limit) or (
                operator is Operator.OFFSET and self.adapter.supports_offset
            ):
                constraints_used.append((filter_index, True))
                filter_index += 1
                indexes.append((LIMIT_OFFSET_INDEX, sqlite_index_constraint))
            # column operator
            elif column_index >= 0:
                column_name = column_names[column_index]
                column_type = column_types[column_index]
                for class_ in column_type.filters:
                    if operator in class_.operators:
                        filtered_columns.append((column_name, operator))
                        constraints_used.append((filter_index, column_type.exact))
                        filter_index += 1
                        indexes.append((column_index, sqlite_index_constraint))
                        break
                else:
                    constraints_used.append(None)

        # estimate query cost
        order = get_order(orderbys, column_names)
        estimated_cost = self.adapter.get_cost(filtered_columns, order)

        # is the data being returned in the requested order? if not, SQLite will have
        # to sort it
        orderby_consumed = True
        orderbys_to_process: list[OrderBy] = []
        for column_index, descending in orderbys:
            requested_order = Order.DESCENDING if descending else Order.ASCENDING
            column_type = column_types[column_index]
            if column_type.order == Order.ANY:
                orderbys_to_process.append((column_index, descending))
            elif column_type.order != requested_order:
                orderby_consumed = False
                break

        return (
            constraints_used,
            index_number,
            indexes,
            orderbys_to_process,
            orderby_consumed,
            estimated_cost,
        )

    def BestIndex(  # pylint: disable=too-many-locals
        self,
        constraints: list[tuple[int, SQLiteConstraint]],
        orderbys: list[OrderBy],
    ) -> tuple[list[Constraint], int, str, bool, float]:
        """
        Build an index for a given set of constraints and order bys.

        The purpose of this method is to ask if you have the ability to determine if
        a row meets certain constraints that doesn’t involve visiting every row.
        """
        (
            constraints_used,
            index_number,
            indexes,
            orderbys_to_process,
            orderby_consumed,
            estimated_cost,
        ) = self._build_index(constraints, orderbys)

        index_name = json.dumps(
            {"indexes": indexes, "orderbys_to_process": orderbys_to_process},
        )

        return (
            constraints_used,
            index_number,
            index_name,
            orderby_consumed,
            estimated_cost,
        )

    def BestIndexObject(  # pylint: disable=too-many-locals
        self,
        index_info: apsw.IndexInfo,
    ) -> bool:
        """
        Alternative to ``BestIndex`` that allows returning only selected columns.
        """
        columns = self.adapter.get_columns()
        column_names = list(columns.keys())

        index_info_dict = index_info_to_dict(index_info, column_names=column_names)
        constraints = [
            (constraint.get("iColumn", -1), constraint["op"])
            for constraint in index_info_dict["aConstraint"]
        ]
        orderbys = [
            (orderby["iColumn"], orderby["desc"])
            for orderby in index_info_dict["aOrderBy"]
        ]
        (
            constraints_used,
            index_number,
            indexes,
            orderbys_to_process,
            orderby_consumed,
            estimated_cost,
        ) = self._build_index(constraints, orderbys)

        requested_columns = sorted(index_info_dict["colUsed_names"])
        index_name = json.dumps(
            {
                "indexes": indexes,
                "orderbys_to_process": orderbys_to_process,
                "requested_columns": requested_columns,
            },
        )

        for i, constraint in enumerate(constraints_used):
            if isinstance(constraint, tuple):
                index_info.set_aConstraintUsage_argvIndex(i, constraint[0] + 1)
                index_info.set_aConstraintUsage_omit(i, constraint[1])
        index_info.idxNum = index_number
        index_info.idxStr = index_name
        index_info.orderByConsumed = orderby_consumed
        index_info.estimatedCost = estimated_cost

        return True

    def Open(self) -> "VTCursor":
        """
        Returns a cursor object.
        """
        return VTCursor(self.adapter)

    def Disconnect(self) -> None:
        """
        The opposite of VTModule.Connect().

        This method is called when a reference to a virtual table is no longer used,
        but VTTable.Destroy() will be called when the table is no longer used.
        """
        self.adapter.close()

    Destroy = Disconnect

    def UpdateInsertRow(self, rowid: Optional[int], fields: tuple[Any, ...]) -> int:
        """
        Insert a row with the specified rowid.
        """
        columns = self.adapter.get_columns()

        row = {column_name: field for field, column_name in zip(fields, columns.keys())}
        row["rowid"] = rowid
        row = next(convert_rows_from_sqlite(columns, iter([row])))

        return cast(int, self.adapter.insert_row(row))

    def UpdateDeleteRow(self, rowid: int) -> None:
        """
        Delete the row with the specified rowid.
        """
        self.adapter.delete_row(rowid)

    def UpdateChangeRow(
        self,
        rowid: int,
        newrowid: int,
        fields: tuple[Any, ...],
    ) -> None:
        """
        Change an existing row.

        Note that the row ID can be modified.
        """
        columns = self.adapter.get_columns()

        row = {column_name: field for field, column_name in zip(fields, columns.keys())}
        row["rowid"] = newrowid
        row = next(convert_rows_from_sqlite(columns, iter([row])))

        self.adapter.update_row(rowid, row)


class VTCursor:
    """
    An object for iterating over a table.
    """

    def __init__(self, adapter: Adapter):
        self.adapter = adapter

        self.data: Iterator[tuple[Any, ...]]
        self.current_row: tuple[Any, ...]
        self.eof = False

    def Filter(  # pylint: disable=too-many-locals
        self,
        indexnumber: int,  # pylint: disable=unused-argument
        indexname: str,
        constraintargs: list[Any],
    ) -> None:
        """
        Filter and sort data according to constraints.

        This method converts the ``indexname`` (containing which columns to filter
        and the order to sort the results) and ``constraintargs`` into a pair of
        ``bounds`` and ``order``. These are then passed to the ``get_rows`` method of
        the adapter, to filter and sort the data.
        """
        columns: dict[str, Field] = self.adapter.get_columns()
        column_names: list[str] = list(columns.keys())
        index = json.loads(indexname)
        indexes: list[Index] = index["indexes"]
        orderbys: list[OrderBy] = index["orderbys_to_process"]

        # compute bounds for each column
        all_bounds = get_all_bounds(indexes, constraintargs, columns)
        limit, offset = get_limit_offset(indexes, constraintargs)
        bounds = get_bounds(columns, all_bounds)

        # compute requested order
        order = get_order(orderbys, column_names)

        # limit and offset were introduced in 1.1, and not all adapters support it
        kwargs: dict[str, Any] = {}
        if self.adapter.supports_limit:
            kwargs["limit"] = limit
        if self.adapter.supports_offset:
            kwargs["offset"] = offset
        if "requested_columns" in index:
            kwargs["requested_columns"] = set(index["requested_columns"])

        rows = self.adapter.get_rows(bounds, order, **kwargs)
        rows = convert_rows_to_sqlite(columns, rows)

        # if a given column is not present, replace it with ``None``
        self.data = (
            tuple(row.get(name) for name in ["rowid", *column_names]) for row in rows
        )
        self.Next()

    def Eof(self) -> bool:
        """
        Called to ask if we are at the end of the table.
        """
        return self.eof

    def Rowid(self) -> int:
        """
        Return the current rowid.
        """
        return cast(int, self.current_row[0])

    def Column(self, col) -> Any:
        """
        Requests the value of the specified column number of the current row.
        """
        return self.current_row[1 + col]

    def Next(self) -> None:
        """
        Move the cursor to the next row.
        """
        try:
            self.current_row = next(self.data)
            self.eof = False
        except StopIteration:
            self.eof = True

    def Close(self) -> None:
        """
        This is the destructor for the cursor.
        """



"""
Fixtures for Shillelagh.
"""

import json
import logging
import os
from collections.abc import Iterator
from typing import Any

import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.registry import AdapterLoader

_logger = logging.getLogger(__name__)


@pytest.fixture
def adapter_kwargs() -> dict[str, Any]:
    """
    Load adapter configuration.

    This fixture looks for the environment variable `SHILLELAGH_ADAPTER_KWARGS`.
    The configuration should be encoded as JSON.
    """
    kwargs = {}
    if "SHILLELAGH_ADAPTER_KWARGS" in os.environ:
        try:
            kwargs = json.loads(os.environ["SHILLELAGH_ADAPTER_KWARGS"])
        except Exception as ex:  # pylint: disable=broad-except
            _logger.warning('Unable to load "SHILLELAGH_ADAPTER_KWARGS": %s', ex)

    return kwargs


@pytest.fixture
def registry(mocker: MockerFixture) -> Iterator[AdapterLoader]:
    """
    Create a custom adapter registry.
    """
    custom_registry = AdapterLoader()
    mocker.patch("shillelagh.adapters.registry.registry", new=custom_registry)
    mocker.patch("shillelagh.backends.apsw.db.registry", new=custom_registry)
    yield custom_registry



"""
A simple example showing the GitHub adapter.
"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = """
SELECT * FROM "https://api.github.com/repos/apache/superset/pulls"
    """
    for row in cursor.execute(SQL):
        print(row)



"""
Test the generic JSON adapter.
"""

import re

import pytest
from requests_mock.mocker import Mocker
from yarl import URL

from shillelagh.adapters.api.generic_json import GenericJSONAPI
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.typing import Maybe

baseurl = URL("https://api.stlouisfed.org/fred/series")


def test_generic_json(requests_mock: Mocker) -> None:
    """
    Test a simple query.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    params = {
        "series_id": "GNPCA",
        "api_key": "abcdefghijklmnopqrstuvwxyz123456",
        "file_type": "json",
    }
    url = (baseurl % params).with_fragment("$.seriess[*]")
    requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(
        str(url),
        json={
            "realtime_start": "2022-11-01",
            "realtime_end": "2022-11-01",
            "seriess": [
                {
                    "id": "GNPCA",
                    "realtime_start": "2022-11-01",
                    "realtime_end": "2022-11-01",
                    "title": "Real Gross National Product",
                    "observation_start": "1929-01-01",
                    "observation_end": "2021-01-01",
                    "frequency": "Annual",
                    "frequency_short": "A",
                    "units": "Billions of Chained 2012 Dollars",
                    "units_short": "Bil. of Chn. 2012 $",
                    "seasonal_adjustment": "Not Seasonally Adjusted",
                    "seasonal_adjustment_short": "NSA",
                    "last_updated": "2022-09-29 07:45:54-05",
                    "popularity": 16,
                    "notes": "BEA Account Code: A001RX\n\n",
                },
            ],
        },
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericjsonapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = f'SELECT * FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [
        (
            "GNPCA",
            "2022-11-01",
            "2022-11-01",
            "Real Gross National Product",
            "1929-01-01",
            "2021-01-01",
            "Annual",
            "A",
            "Billions of Chained 2012 Dollars",
            "Bil. of Chn. 2012 $",
            "Not Seasonally Adjusted",
            "NSA",
            "2022-09-29 07:45:54-05",
            16,
            "BEA Account Code: A001RX\n\n",
        ),
    ]

    requests_mock.head(
        "https://example.org/data.json",
        headers={"content-type": "application/json"},
    )
    requests_mock.get(
        "https://example.org/data.json",
        json=[{"a": 1, "b": [10, 20]}, {"a": 2, "b": [11]}],
    )
    sql = 'SELECT a, b FROM "https://example.org/data.json"'
    rows = list(cursor.execute(sql))
    assert rows == [(1, "[10, 20]"), (2, "[11]")]

    requests_mock.get(
        "https://example.org/data.json",
        json={
            "error": {"code": 1002, "message": "API key is invalid or not provided."},
        },
        status_code=500,
    )
    with pytest.raises(ProgrammingError) as excinfo:
        list(cursor.execute(sql))
    assert str(excinfo.value) == "Error: API key is invalid or not provided."


def test_generic_json_complex_type(requests_mock: Mocker) -> None:
    """
    Test a query where columns are complex.
    """
    # for datassette and other probing adapters
    requests_mock.head("https://example.org/-/versions.json", status_code=404)

    url = URL("https://example.org/")
    requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(
        str(url),
        json=[
            {
                "foo": "bar",
                "baz": ["one", "two"],
            },
        ],
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericjsonapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = f'SELECT * FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [("bar", '["one", "two"]')]


def test_supports(requests_mock: Mocker) -> None:
    """
    Test the ``supports`` method.
    """
    requests_mock.head(
        "https://example.org/data.html",
        headers={"content-type": "text/html"},
    )
    requests_mock.head(
        "https://example.org/data.json",
        headers={"content-type": "application/json"},
    )

    assert GenericJSONAPI.supports("/etc/password") is False
    assert GenericJSONAPI.supports("https://example.org/data.html") is Maybe
    assert GenericJSONAPI.supports("https://example.org/data.html", fast=False) is False
    assert GenericJSONAPI.supports("https://example.org/data.json", fast=False) is True


def test_request_headers(requests_mock: Mocker) -> None:
    """
    Test passing requests headers.
    """
    supports = requests_mock.head(
        "https://example.org/data.json",
        headers={"content-type": "application/json"},
    )

    # for datassette and other probing adapters
    requests_mock.head("https://example.org/-/versions.json", status_code=404)

    url = URL("https://example.org/")
    data = requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(
        str(url),
        json=[
            {
                "foo": "bar",
                "baz": ["one", "two"],
            },
        ],
    )

    # test the supports method
    GenericJSONAPI.supports(
        "https://example.org/data.json",
        fast=False,
        request_headers={"foo": "bar"},
        cache_expiration=-1,
    )
    assert supports.last_request.headers["foo"] == "bar"

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "genericjsonapi": {
                "request_headers": {"foo": "bar"},
                "cache_expiration": -1,
            },
        },
    )
    cursor = connection.cursor()

    sql = f'SELECT * FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [("bar", '["one", "two"]')]
    assert data.last_request.headers["foo"] == "bar"


def test_request_headers_in_url(requests_mock: Mocker) -> None:
    """
    Test passing requests headers.
    """
    supports = requests_mock.head(
        "https://example.org/data.json",
        headers={"content-type": "application/json"},
    )

    # for datassette and other probing adapters
    requests_mock.head("https://example.org/-/versions.json", status_code=404)

    url = URL("https://example.org/")
    data = requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(
        str(url),
        json=[
            {
                "foo": "bar",
                "baz": ["one", "two"],
            },
        ],
    )

    # test the supports method
    GenericJSONAPI.supports(
        "https://example.org/data.json?_s_headers=(foo:bar)",
        fast=False,
        cache_expiration=-1,
    )
    assert supports.last_request.headers["foo"] == "bar"

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericjsonapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = 'SELECT * FROM "https://example.org/?_s_headers=(foo:bar)"'
    rows = list(cursor.execute(sql))
    assert rows == [("bar", '["one", "two"]')]
    assert data.last_request.headers["foo"] == "bar"


def test_single_row(requests_mock: Mocker) -> None:
    """
    Test a query where the response is a single row as a dictionary.
    """
    # for datassette
    requests_mock.get(re.compile(".*-/versions.json.*"), status_code=404)

    url = "https://www.boredapi.com/api/activity?participants=1#$"
    requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(
        str(url),
        json={
            "activity": "Solve a Rubik's cube",
            "type": "recreational",
            "participants": 1,
            "price": 0,
            "link": "",
            "key": "4151544",
            "accessibility": 0.1,
        },
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericjsonapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = f'SELECT * FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [("Solve a Rubik's cube", "recreational", 1, 0, "", "4151544", 0.1)]


def test_generic_json_array(requests_mock: Mocker) -> None:
    """
    Test a query where the response has only arrays.
    """
    # for datassette and other probing adapters
    requests_mock.get(
        "https://example.com/-/versions.json",
        status_code=404,
    )

    url = URL("https://example.com/")
    requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(
        str(url),
        json=[
            [0, 0, 15],
            [0, 1, 8],
            [0, 2, 6],
            [0, 3, 3],
            [0, 4, 2],
            None,
        ],
    )

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericjsonapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = f'SELECT * FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [
        (0, 0, 15),
        (0, 1, 8),
        (0, 2, 6),
        (0, 3, 3),
        (0, 4, 2),
        (None, None, None),
    ]
    assert cursor.description
    assert {t[0] for t in cursor.description} == {"col_0", "col_1", "col_2"}


def test_generic_json_string_array(requests_mock: Mocker) -> None:
    """
    Test a query where the JSONPath returns just a string.
    """
    payload = {
        "lastTimestamp": "20240808022653Z",
        "project_count": 243,
        "roster_counts": {
            "superset": {"members": 65, "owners": 35},
        },
        "projects": {
            "superset": {
                "createTimestamp": "20170523163111Z",
                "modifyTimestamp": "20240617171611Z",
                "member_count": 65,
                "owner_count": 35,
                "members": [
                    "aafghahi",
                    "amitmiran",
                    "arivero",
                    "benceorlai",
                    "beto",
                ],
                "owners": [
                    "amitmiran",
                    "beto",
                    "bkyryliuk",
                    "ccwilliams",
                    "craigrueda",
                ],
                "pmc": True,
                "podling": "graduated",
            },
        },
    }
    url = URL("https://example.com/#$.projects.superset.members[*]")
    requests_mock.head(str(url), headers={"content-type": "application/json"})
    requests_mock.get(str(url), json=payload)

    connection = connect(
        ":memory:",
        adapter_kwargs={"genericjsonapi": {"cache_expiration": -1}},
    )
    cursor = connection.cursor()

    sql = f'SELECT * FROM "{url}"'
    rows = list(cursor.execute(sql))
    assert rows == [
        ("aafghahi",),
        ("amitmiran",),
        ("arivero",),
        ("benceorlai",),
        ("beto",),
    ]

    assert cursor.description
    assert {t[0] for t in cursor.description} == {"col_0"}



"""DB API 2.0 types for Shillelagh."""

import datetime
import inspect
import time
from typing import Any

from shillelagh.fields import Field


class DBAPIType:  # pylint: disable=too-few-public-methods
    """
    Constructor for the required DB API 2.0 types.
    """

    def __init__(self, name: str):
        self.name = name

    def __eq__(self, other: Any) -> bool:
        if inspect.isclass(other) and issubclass(other, Field):
            return bool(self.name == other.db_api_type)

        return NotImplemented


STRING = DBAPIType("STRING")
BINARY = DBAPIType("BINARY")
NUMBER = DBAPIType("NUMBER")
DATETIME = DBAPIType("DATETIME")
ROWID = DBAPIType("ROWID")


def Date(  # pylint: disable=invalid-name
    year: int,
    month: int,
    day: int,
) -> datetime.date:
    """Constructs an object holding a date value."""
    return datetime.date(year, month, day)


def Time(  # pylint: disable=invalid-name
    hour: int,
    minute: int,
    second: int,
) -> datetime.time:
    """Constructs an object holding a time value."""
    return datetime.time(hour, minute, second, tzinfo=datetime.timezone.utc)


def Timestamp(  # pylint: disable=invalid-name, too-many-arguments, too-many-positional-arguments
    year: int,
    month: int,
    day: int,
    hour: int,
    minute: int,
    second: int,
) -> datetime.datetime:
    """Constructs an object holding a timestamp value."""
    return datetime.datetime(
        year,
        month,
        day,
        hour,
        minute,
        second,
        tzinfo=datetime.timezone.utc,
    )


def DateFromTicks(ticks: int) -> datetime.date:  # pylint: disable=invalid-name
    """
    Constructs an object holding a date value from the given ticks value.

    Ticks should be in number of seconds since the epoch.
    """
    return Date(*time.gmtime(ticks)[:3])


def TimeFromTicks(ticks: int) -> datetime.time:  # pylint: disable=invalid-name
    """
    Constructs an object holding a time value from the given ticks value.

    Ticks should be in number of seconds since the epoch.
    """
    return Time(*time.gmtime(ticks)[3:6])


def TimestampFromTicks(ticks: int) -> datetime.datetime:  # pylint: disable=invalid-name
    """
    Constructs an object holding a timestamp value from the given ticks value.

    Ticks should be in number of seconds since the epoch.
    """
    return Timestamp(*time.gmtime(ticks)[:6])


def Binary(string: str) -> bytes:  # pylint: disable=invalid-name
    """constructs an object capable of holding a binary (long) string value."""
    return string.encode()



"""
A SQLALchemy dialect.
"""

# pylint: disable=protected-access, abstract-method

from typing import Any, Optional, cast

import sqlalchemy.types
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.engine.url import URL
from sqlalchemy.pool.base import _ConnectionFairy
from sqlalchemy.sql.type_api import TypeEngine
from typing_extensions import TypedDict

from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw import db
from shillelagh.backends.apsw.vt import VTTable
from shillelagh.exceptions import ProgrammingError
from shillelagh.lib import find_adapter


class SQLAlchemyColumn(TypedDict):
    """
    A custom type for a SQLAlchemy column.
    """

    name: str
    type: TypeEngine
    nullable: bool
    default: Optional[str]
    autoincrement: str
    primary_key: int


class APSWDialect(SQLiteDialect):
    """
    A SQLAlchemy dialect for Shillelagh.

    The dialect is based on the ``SQLiteDialect``, since we're using APSW.
    """

    name = "shillelagh"
    driver = "apsw"

    # This is supported in ``SQLiteDialect``, and equally supported here. See
    # https://docs.sqlalchemy.org/en/14/core/connections.html#caching-for-third-party-dialects
    # for more context.
    supports_statement_cache = True

    # ``SQLiteDialect.colspecs`` has custom representations for objects that SQLite stores
    # as string (eg, timestamps). Since the Shillelagh DB API driver returns them as
    # proper objects the custom representations are not needed.
    colspecs: dict[TypeEngine, TypeEngine] = {}

    supports_sane_rowcount = False

    @classmethod
    def dbapi(cls):  # pylint: disable=method-hidden
        """
        Return the DB API module.
        """
        return db

    @classmethod
    def import_dbapi(cls):  # pylint: disable=method-hidden
        """
        Return the DB API module.
        """
        return db

    def __init__(
        self,
        adapters: Optional[list[str]] = None,
        adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
        safe: bool = False,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs or {}
        self._safe = safe

    def create_connect_args(
        self,
        url: URL,
    ) -> tuple[tuple[()], dict[str, Any]]:
        path = str(url.database) if url.database else ":memory:"
        return (), {
            "path": path,
            "adapters": self._adapters,
            "adapter_kwargs": self._adapter_kwargs,
            "safe": self._safe,
            "isolation_level": self.isolation_level,
        }

    def do_ping(self, dbapi_connection: _ConnectionFairy) -> bool:
        """
        Return true if the database is online.
        """
        return True

    def has_table(  # pylint: disable=unused-argument
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> bool:
        """
        Return true if a given table exists.
        """
        try:
            get_adapter_for_table_name(connection, table_name)
        except ProgrammingError:
            return bool(
                super().has_table(
                    connection,
                    table_name,
                    schema,
                    **kwargs,  # pylint: disable=unused-argument
                ),
            )
        return True

    # needed for SQLAlchemy
    def _get_table_sql(  # pylint: disable=unused-argument
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        adapter = get_adapter_for_table_name(connection, table_name)
        table = VTTable(adapter)
        return table.get_create_table(table_name)

    def get_columns(  # pylint: disable=unused-argument
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> list[SQLAlchemyColumn]:
        adapter = get_adapter_for_table_name(connection, table_name)
        columns = adapter.get_columns()
        return [
            {
                "name": column_name,
                "type": getattr(sqlalchemy.types, field.type),
                "nullable": True,
                "default": None,
                "autoincrement": "auto",
                "primary_key": 0,
            }
            for column_name, field in columns.items()
        ]


def get_adapter_for_table_name(
    connection: _ConnectionFairy,
    table_name: str,
) -> Adapter:
    """
    Return an adapter associated with a connection.

    This function instantiates the adapter responsible for a given table name,
    using the connection to properly pass any adapter kwargs.
    """
    raw_connection = cast(db.Connection, connection.engine.raw_connection())
    adapter, args, kwargs = find_adapter(
        table_name,
        raw_connection._adapter_kwargs,
        raw_connection._adapters,
    )
    return adapter(*args, **kwargs)  # type: ignore



"""
Tests for shillelagh.backends.apsw.dialects.safe.
"""

from sqlalchemy.engine.url import make_url

from shillelagh.backends.apsw.dialects.safe import APSWSafeDialect


def test_safe_dialect() -> None:
    """
    Test that ``shillelagh+safe://`` forces safe mode.
    """
    dialect = APSWSafeDialect()
    assert dialect.create_connect_args(make_url("shillelagh+safe://")) == (
        (),
        {
            "path": ":memory:",
            "adapters": None,
            "adapter_kwargs": {},
            "safe": True,
            "isolation_level": None,
        },
    )



"""
Tests for the Multicorn2 FDW.
"""

# pylint: disable=invalid-name, redefined-outer-name, no-member, redefined-builtin

from collections import defaultdict

from multicorn import Qual, SortKey
from pytest_mock import MockerFixture

from shillelagh.adapters.registry import AdapterLoader
from shillelagh.backends.multicorn.fdw import (
    MulticornForeignDataWrapper,
    get_all_bounds,
)
from shillelagh.filters import Operator

from ...fakes import FakeAdapter


def test_fdw(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``MulticornForeignDataWrapper`` class.
    """
    mocker.patch("shillelagh.backends.multicorn.fdw.registry", registry)

    registry.add("dummy", FakeAdapter)

    assert (
        MulticornForeignDataWrapper.import_schema("schema", {}, {}, "limit", []) == []
    )
    wrapper = MulticornForeignDataWrapper(
        {"adapter": "dummy", "args": "qQA="},
        {},
    )
    assert wrapper.rowid_column == "rowid"

    assert list(wrapper.execute([], ["rowid", "name", "age", "pets"])) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
    ]

    assert list(
        wrapper.execute(
            [Qual("age", ">", 21)],
            ["rowid", "name", "age", "pets"],
            [],
        ),
    ) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
    ]

    assert list(
        wrapper.execute(
            [],
            ["rowid", "name", "age", "pets"],
            [
                SortKey(
                    attname="age",
                    attnum=2,
                    is_reversed=True,
                    nulls_first=True,
                    collate=None,
                ),
            ],
        ),
    ) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
    ]


def test_get_all_bounds() -> None:
    """
    Test ``get_all_bounds``.
    """
    quals = [
        Qual("column1", "=", 3),
        Qual("column2", "LIKE", "test%"),
        Qual("column3", ">", 10),
    ]

    assert get_all_bounds([]) == defaultdict(set)
    assert get_all_bounds([quals[0]]) == {"column1": {(Operator.EQ, 3)}}
    assert get_all_bounds(quals) == {
        "column1": {(Operator.EQ, 3)},
        "column3": {(Operator.GT, 10)},
    }
    assert get_all_bounds([Qual("column4", "unsupported_operator", 1)]) == defaultdict(
        set,
    )


def test_can_sort(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``can_sort`` method.
    """
    mocker.patch("shillelagh.backends.multicorn.fdw.registry", registry)

    registry.add("dummy", FakeAdapter)

    wrapper = MulticornForeignDataWrapper(
        {"adapter": "dummy", "args": "qQA="},
        {},
    )
    assert wrapper.can_sort([]) == []
    assert wrapper.can_sort(
        [
            SortKey(
                attname="age",
                attnum=2,
                is_reversed=True,
                nulls_first=True,
                collate=None,
            ),
            SortKey(
                attname="foobar",
                attnum=1,
                is_reversed=True,
                nulls_first=True,
                collate=None,
            ),
        ],
    ) == [
        SortKey(
            attname="age",
            attnum=2,
            is_reversed=True,
            nulls_first=True,
            collate=None,
        ),
    ]


def test_insert(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``insert`` method.
    """
    mocker.patch("shillelagh.backends.multicorn.fdw.registry", registry)

    registry.add("dummy", FakeAdapter)

    wrapper = MulticornForeignDataWrapper(
        {"adapter": "dummy", "args": "qQA="},
        {},
    )

    wrapper.insert({"rowid": 2, "name": "Charlie", "age": 6, "pets": 1})
    assert list(wrapper.execute([], ["rowid", "name", "age", "pets"])) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 2, "name": "Charlie", "age": 6, "pets": 1},
    ]


def test_delete(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``delete`` method.
    """
    mocker.patch("shillelagh.backends.multicorn.fdw.registry", registry)

    registry.add("dummy", FakeAdapter)

    wrapper = MulticornForeignDataWrapper(
        {"adapter": "dummy", "args": "qQA="},
        {},
    )

    wrapper.delete({"rowid": 1, "name": "Bob", "age": 23, "pets": 3})
    assert list(wrapper.execute([], ["rowid", "name", "age", "pets"])) == [
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
    ]


def test_update(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``update`` method.
    """
    mocker.patch("shillelagh.backends.multicorn.fdw.registry", registry)

    registry.add("dummy", FakeAdapter)

    wrapper = MulticornForeignDataWrapper(
        {"adapter": "dummy", "args": "qQA="},
        {},
    )

    wrapper.update(
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 0},
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 1},
    )
    assert list(wrapper.execute([], ["rowid", "name", "age", "pets"])) == [
        {"rowid": 1, "name": "Bob", "age": 23, "pets": 3},
        {"rowid": 0, "name": "Alice", "age": 20, "pets": 1},
    ]


def test_get_rel_Size(mocker: MockerFixture, registry: AdapterLoader) -> None:
    """
    Test the ``get_rel_size`` method.
    """
    mocker.patch("shillelagh.backends.multicorn.fdw.registry", registry)

    registry.add("dummy", FakeAdapter)

    wrapper = MulticornForeignDataWrapper(
        {"adapter": "dummy", "args": "qQA="},
        {},
    )

    assert wrapper.get_rel_size([Qual("age", ">", 21)], ["name", "age"]) == (666, 200)



"""
Tests for the DbtMetricFlowAPI adapter.
"""

# pylint: disable=line-too-long, invalid-name, unused-argument, redefined-outer-name, protected-access, too-many-lines

import base64
from datetime import datetime, timezone
from decimal import Decimal
from typing import Any
from unittest.mock import MagicMock, call

import pandas as pd
import pyarrow as pa
import pytest
from pytest_mock import MockerFixture

from shillelagh.adapters.api.dbt_metricflow import (
    DbtMetricFlowAPI,
    Timestamp,
    find_cursor,
)
from shillelagh.backends.apsw.db import Cursor, connect
from shillelagh.exceptions import ImpossibleFilterError, InternalError, ProgrammingError
from shillelagh.fields import Boolean, Date
from shillelagh.fields import Decimal as DecimalField
from shillelagh.fields import Field, Integer, Order, String, Time, Unknown
from shillelagh.filters import (
    Equal,
    Filter,
    Impossible,
    IsNotNull,
    IsNull,
    NotEqual,
    Range,
)

INIT_PAYLOAD = [
    {
        "data": {
            "metrics": [
                {
                    "name": "orders",
                    "description": "Count of orders.",
                    "type": "SIMPLE",
                    "dimensions": [
                        {
                            "name": "order_id__is_food_order",
                            "description": (
                                "A boolean indicating if this order included any food "
                                "items."
                            ),
                            "queryableGranularities": [],
                            "type": "CATEGORICAL",
                        },
                    ],
                },
            ],
        },
    },
    {"data": {"createQuery": {"queryId": "2fWo7hWqRbyMJpBbBNTwD6WX2jb"}}},
    {
        "data": {
            "query": {
                "sql": """SELECT
  is_food_order AS order_id__is_food_order
FROM `dbt-tutorial-347100`.`dbt_beto`.`orders` orders_src_10000
GROUP BY
  order_id__is_food_order""",
                "status": "SUCCESSFUL",
                "error": None,
                "arrowResult": "/////4AAAAAQAAAAAAAKAAwABgAFAAgACgAAAAABBAAMAAAACAAIAAAABAAIAAAABAAAAAEAAAAUAAAAEAAUAAgABgAHAAwAAAAQABAAAAAAAAEGEAAAACwAAAAEAAAAAAAAABcAAABvcmRlcl9pZF9faXNfZm9vZF9vcmRlcgAEAAQABAAAAP////+IAAAAFAAAAAAAAAAMABYABgAFAAgADAAMAAAAAAMEABgAAAAQAAAAAAAAAAAACgAYAAwABAAIAAoAAAA8AAAAEAAAAAMAAAAAAAAAAAAAAAIAAAAAAAAAAAAAAAEAAAAAAAAACAAAAAAAAAABAAAAAAAAAAAAAAABAAAAAwAAAAAAAAABAAAAAAAAAAYAAAAAAAAAAgAAAAAAAAD/////AAAAAA==",
            },
        },
    },
]


@pytest.fixture
def client(mocker: MockerFixture) -> MagicMock:
    """
    A client that mocks a successful query.
    """
    GraphqlClient: MagicMock = mocker.patch(
        "shillelagh.adapters.api.dbt_metricflow.GraphqlClient",
    )
    GraphqlClient().execute.side_effect = [
        *INIT_PAYLOAD,
        {"data": {"createQuery": {"queryId": "2fThegMTT9iBqmUHQE0gXY7dzEJ"}}},
        {
            "data": {
                "query": {
                    "sql": """SELECT
  order_id__is_food_order
  , SUM(order_count) AS orders
FROM (
  SELECT
    is_food_order AS order_id__is_food_order
    , 1 AS order_count
  FROM `dbt-tutorial-347100`.`dbt_beto`.`orders` orders_src_10000
) subq_2
GROUP BY
  order_id__is_food_order
LIMIT 10""",
                    "status": "COMPILED",
                    "error": None,
                    "arrowResult": None,
                },
            },
        },
        {
            "data": {
                "query": {
                    "sql": """SELECT
  order_id__is_food_order
  , SUM(order_count) AS orders
FROM (
  SELECT
    is_food_order AS order_id__is_food_order
    , 1 AS order_count
  FROM `dbt-tutorial-347100`.`dbt_beto`.`orders` orders_src_10000
) subq_2
GROUP BY
  order_id__is_food_order
LIMIT 10""",
                    "status": "COMPILED",
                    "error": None,
                    "arrowResult": "/////7gAAAAQAAAAAAAKAAwABgAFAAgACgAAAAABBAAMAAAACAAIAAAABAAIAAAABAAAAAIAAABMAAAABAAAAMz///8AAAEHEAAAACAAAAAEAAAAAAAAAAYAAABvcmRlcnMAAAAABgAIAAQABgAAABMAAAAQABQACAAGAAcADAAAABAAEAAAAAAAAQYQAAAALAAAAAQAAAAAAAAAFwAAAG9yZGVyX2lkX19pc19mb29kX29yZGVyAAQABAAEAAAA/////7gAAAAUAAAAAAAAAAwAFgAGAAUACAAMAAwAAAAAAwQAGAAAAEAAAAAAAAAAAAAKABgADAAEAAgACgAAAFwAAAAQAAAAAwAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAQAAAAAAAAAIAAAAAAAAAAEAAAAAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAMAAAAAAAAAAAAAAAAgAAAAMAAAAAAAAAAQAAAAAAAAADAAAAAAAAAAAAAAAAAAAABgAAAAAAAAACAAAAAAAAAF8EAAAAAAAAAAAAAAAAAABVRwAAAAAAAAAAAAAAAAAAUJ0AAAAAAAAAAAAAAAAAAP////8AAAAA",
                },
            },
        },
        {
            "data": {
                "query": {
                    "sql": """SELECT
  order_id__is_food_order
  , SUM(order_count) AS orders
FROM (
  SELECT
    is_food_order AS order_id__is_food_order
    , 1 AS order_count
  FROM `dbt-tutorial-347100`.`dbt_beto`.`orders` orders_src_10000
) subq_2
GROUP BY
  order_id__is_food_order
LIMIT 10""",
                    "status": "SUCCESSFUL",
                    "error": None,
                    "arrowResult": "/////7gAAAAQAAAAAAAKAAwABgAFAAgACgAAAAABBAAMAAAACAAIAAAABAAIAAAABAAAAAIAAABMAAAABAAAAMz///8AAAEHEAAAACAAAAAEAAAAAAAAAAYAAABvcmRlcnMAAAAABgAIAAQABgAAABMAAAAQABQACAAGAAcADAAAABAAEAAAAAAAAQYQAAAALAAAAAQAAAAAAAAAFwAAAG9yZGVyX2lkX19pc19mb29kX29yZGVyAAQABAAEAAAA/////7gAAAAUAAAAAAAAAAwAFgAGAAUACAAMAAwAAAAAAwQAGAAAAEAAAAAAAAAAAAAKABgADAAEAAgACgAAAFwAAAAQAAAAAwAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAQAAAAAAAAAIAAAAAAAAAAEAAAAAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAMAAAAAAAAAAAAAAAAgAAAAMAAAAAAAAAAQAAAAAAAAADAAAAAAAAAAAAAAAAAAAABgAAAAAAAAACAAAAAAAAAF8EAAAAAAAAAAAAAAAAAABVRwAAAAAAAAAAAAAAAAAAUJ0AAAAAAAAAAAAAAAAAAP////8AAAAA",
                },
            },
        },
        {
            "data": {
                "query": {
                    "sql": """SELECT
  order_id__is_food_order
  , SUM(order_count) AS orders
FROM (
  SELECT
    is_food_order AS order_id__is_food_order
    , 1 AS order_count
  FROM `dbt-tutorial-347100`.`dbt_beto`.`orders` orders_src_10000
) subq_2
GROUP BY
  order_id__is_food_order
LIMIT 10""",
                    "status": "SUCCESSFUL",
                    "error": None,
                    "arrowResult": "/////7gAAAAQAAAAAAAKAAwABgAFAAgACgAAAAABBAAMAAAACAAIAAAABAAIAAAABAAAAAIAAABMAAAABAAAAMz///8AAAEHEAAAACAAAAAEAAAAAAAAAAYAAABvcmRlcnMAAAAABgAIAAQABgAAABMAAAAQABQACAAGAAcADAAAABAAEAAAAAAAAQYQAAAALAAAAAQAAAAAAAAAFwAAAG9yZGVyX2lkX19pc19mb29kX29yZGVyAAQABAAEAAAA/////7gAAAAUAAAAAAAAAAwAFgAGAAUACAAMAAwAAAAAAwQAGAAAAEAAAAAAAAAAAAAKABgADAAEAAgACgAAAFwAAAAQAAAAAwAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAQAAAAAAAAAIAAAAAAAAAAEAAAAAAAAAEAAAAAAAAAAAAAAAAAAAABAAAAAAAAAAMAAAAAAAAAAAAAAAAgAAAAMAAAAAAAAAAQAAAAAAAAADAAAAAAAAAAAAAAAAAAAABgAAAAAAAAACAAAAAAAAAF8EAAAAAAAAAAAAAAAAAABVRwAAAAAAAAAAAAAAAAAAUJ0AAAAAAAAAAAAAAAAAAP////8AAAAA",
                },
            },
        },
    ]
    return GraphqlClient


@pytest.fixture
def client_with_error(mocker: MockerFixture) -> None:
    """
    A client that mocks a query with an error.
    """
    GraphqlClient = mocker.patch("shillelagh.adapters.api.dbt_metricflow.GraphqlClient")
    GraphqlClient().execute.side_effect = [
        *INIT_PAYLOAD,
        {"data": {"createQuery": {"queryId": "2fThegMTT9iBqmUHQE0gXY7dzEJ"}}},
        {
            "data": {
                "query": {
                    "sql": """SELECT
  customers_src_10000.customer_name AS customer__customer_name
  , COUNT(DISTINCT orders_src_10000.customer_id) AS customers_with_orders
FROM `dbt-tutorial-347100`.`dbt_beto`.`orders` orders_src_10000
LEFT OUTER JOIN
  `dbt-tutorial-347100`.`dbt_beto`.`customers` customers_src_10000
ON
  orders_src_10000.customer_id = customers_src_10000.customer_id
GROUP BY
  customer__customer_name
LIMIT 10""",
                    "status": "FAILED",
                    "error": """INVALID_ARGUMENT: [FlightSQL] [Simba][BigQueryJDBCDriver](100032) Error executing query job. Message: 400 Bad Request
POST https://bigquery.googleapis.com/bigquery/v2/projects/dbt-tutorial-347100/jobs
{
  "code": 400,
  "errors": [
    {
      "domain": "global",
      "location": "q",
      "locationType": "parameter",
      "message": "No matching signature for operator = for argument types: STRING, INT64. Supported signature: ANY = ANY at [8:3]",
      "reason": "invalidQuery"
    }
  ],
  "message": "No matching signature for operator = for argument types: STRING, INT64. Supported signature: ANY = ANY at [8:3]",
  "status": "INVALID_ARGUMENT"
} (InvalidArgument; Prepare)""",
                    "arrowResult": None,
                },
            },
        },
    ]


@pytest.fixture
def client_with_bad_request(mocker: MockerFixture) -> None:
    """
    A client that does an incorrect request when creating a query.
    """
    GraphqlClient = mocker.patch("shillelagh.adapters.api.dbt_metricflow.GraphqlClient")
    GraphqlClient().execute.side_effect = [
        {
            "data": {
                "metrics": [
                    {
                        "name": "orders",
                        "description": "Count of orders.",
                        "type": "SIMPLE",
                        "dimensions": [
                            {
                                "name": "order_id__is_food_order",
                                "description": (
                                    "A boolean indicating if this order included any "
                                    "food items."
                                ),
                                "queryableGranularities": [],
                                "type": "CATEGORICAL",
                            },
                        ],
                    },
                ],
            },
        },
        {
            "data": None,
            "errors": [
                {
                    "message": "Variable '$environmentId' of required type 'BigInt!' was not provided.",
                    "locations": [{"line": 3, "column": 9}],
                },
            ],
        },
    ]


def test_dbtmetricflowapi(mocker: MockerFixture, client: MockerFixture) -> None:
    """
    Run SQL against the adapter.
    """
    mocker.patch("time.sleep")

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "dbtmetricflowapi": {
                "service_token": "dbtc_XXX",
                "environment_id": 123456,
            },
        },
    )
    cursor = connection.cursor()

    sql = """
        SELECT orders, order_id__is_food_order
        FROM "https://semantic-layer.cloud.getdbt.com/"
        LIMIT 10
    """
    cursor.execute(sql)
    assert cursor.fetchall() == [
        (Decimal("1119"), None),
        (Decimal("18261"), 1),
        (Decimal("40272"), 0),
    ]

    client().execute.assert_has_calls(
        [
            call(
                query="""
    query GetMetrics($environmentId: BigInt!) {
        metrics(environmentId: $environmentId) {
            name
            description
            type
            dimensions {
                name
                description
                queryableGranularities
                type
                expr
            }
        }
    }
""",
                variables={"environmentId": 123456},
            ),
            call(
                query="""
    mutation GetData(
        $environmentId: BigInt!,
        $metrics: [MetricInput!]!,
        $where: [WhereInput!]!,
        $groupBy: [GroupByInput!]!,
        $orderBy: [OrderByInput!]!,
        $limit: Int,
    ) {
        createQuery(
            environmentId: $environmentId,
            metrics: $metrics,
            where: $where,
            groupBy: $groupBy,
            orderBy: $orderBy,
            limit: $limit,
        ) {
            queryId
        }
    }
""",
                variables={
                    "environmentId": 123456,
                    "metrics": [],
                    "where": [],
                    "groupBy": [{"name": "order_id__is_food_order"}],
                    "orderBy": [],
                    "limit": 0,
                },
            ),
            call(
                query="""
    query PollResults(
        $environmentId: BigInt!,
        $queryId: String!,
    ) {
        query(
            environmentId: $environmentId,
            queryId: $queryId,
        ) {
            sql
            status
            error
            arrowResult
        }
    }
""",
                variables={
                    "environmentId": 123456,
                    "queryId": "2fWo7hWqRbyMJpBbBNTwD6WX2jb",
                },
            ),
            call(
                query="""
    mutation GetData(
        $environmentId: BigInt!,
        $metrics: [MetricInput!]!,
        $where: [WhereInput!]!,
        $groupBy: [GroupByInput!]!,
        $orderBy: [OrderByInput!]!,
        $limit: Int,
    ) {
        createQuery(
            environmentId: $environmentId,
            metrics: $metrics,
            where: $where,
            groupBy: $groupBy,
            orderBy: $orderBy,
            limit: $limit,
        ) {
            queryId
        }
    }
""",
                variables={
                    "environmentId": 123456,
                    "metrics": [{"name": "orders"}],
                    "where": [],
                    "groupBy": [{"name": "order_id__is_food_order"}],
                    "orderBy": [],
                    "limit": 10,
                },
            ),
            call(
                query="""
    query PollResults(
        $environmentId: BigInt!,
        $queryId: String!,
    ) {
        query(
            environmentId: $environmentId,
            queryId: $queryId,
        ) {
            sql
            status
            error
            arrowResult
        }
    }
""",
                variables={
                    "environmentId": 123456,
                    "queryId": "2fThegMTT9iBqmUHQE0gXY7dzEJ",
                },
            ),
            call(
                query="""
    query PollResults(
        $environmentId: BigInt!,
        $queryId: String!,
    ) {
        query(
            environmentId: $environmentId,
            queryId: $queryId,
        ) {
            sql
            status
            error
            arrowResult
        }
    }
""",
                variables={
                    "environmentId": 123456,
                    "queryId": "2fThegMTT9iBqmUHQE0gXY7dzEJ",
                },
            ),
            call(
                query="""
    query PollResults(
        $environmentId: BigInt!,
        $queryId: String!,
    ) {
        query(
            environmentId: $environmentId,
            queryId: $queryId,
        ) {
            sql
            status
            error
            arrowResult
        }
    }
""",
                variables={
                    "environmentId": 123456,
                    "queryId": "2fThegMTT9iBqmUHQE0gXY7dzEJ",
                },
            ),
        ],
    )


def test_dbtmetricflowapi_error(
    mocker: MockerFixture,
    client_with_error: MockerFixture,
) -> None:
    """
    Test error capturing.
    """
    mocker.patch("time.sleep")

    connection = connect(
        ":memory:",
        adapter_kwargs={
            "dbtmetricflowapi": {
                "service_token": "dbtc_XXX",
                "environment_id": 123456,
            },
        },
    )
    cursor = connection.cursor()

    sql = """
        SELECT order_id__is_food_order
        FROM "https://semantic-layer.cloud.getdbt.com/"
        LIMIT 10
    """
    with pytest.raises(ProgrammingError):
        cursor.execute(sql)


def test_get_data_requested_columns(
    mocker: MockerFixture,
    client: MockerFixture,
) -> None:
    """
    Test ``get_data`` with newer versions of apsw with ``requested_columns``.
    """
    cursor = mocker.MagicMock()
    cursor.operation = "SELECT orders, order_id__is_food_order FROM t"
    mocker.patch(
        "shillelagh.adapters.api.dbt_metricflow.find_cursor",
        return_value=cursor,
    )
    mocker.patch("time.sleep")

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )
    assert list(adapter.get_data(bounds={}, order=[])) == [
        {"rowid": 0, "order_id__is_food_order": None, "orders": Decimal("1119")},
        {"rowid": 1, "order_id__is_food_order": True, "orders": Decimal("18261")},
        {"rowid": 2, "order_id__is_food_order": False, "orders": Decimal("40272")},
    ]


def test_get_data_no_cursor(
    mocker: MockerFixture,
    client: MockerFixture,
) -> None:
    """
    Test that ``get_data`` fails when we can't fnd the cursor.
    """
    mocker.patch(
        "shillelagh.adapters.api.dbt_metricflow.find_cursor",
        return_value=None,
    )

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )
    with pytest.raises(InternalError) as excinfo:
        list(adapter.get_data(bounds={}, order=[]))
    assert str(excinfo.value) == "Unable to get reference to cursor"


def test_parse_uri() -> None:
    """
    Test the ``parse_uri`` method.
    """
    assert DbtMetricFlowAPI.parse_uri("https://semantic-layer.cloud.getdbt.com/") == (
        "https://semantic-layer.cloud.getdbt.com/",
    )
    assert DbtMetricFlowAPI.parse_uri("https://ab123.us1.dbt.com/") == (
        "https://ab123.us1.dbt.com/",
    )


def test_get_endpoint() -> None:
    """
    Test the ``_get_endpoint`` method.
    """
    assert (
        DbtMetricFlowAPI._get_endpoint("https://semantic-layer.cloud.getdbt.com/")
        == "https://semantic-layer.cloud.getdbt.com/api/graphql"
    )
    assert DbtMetricFlowAPI._get_endpoint("https://ab123.us1.dbt.com/") == (
        "https://ab123.semantic-layer.us1.dbt.com/api/graphql"
    )


def test_build_where(
    mocker: MockerFixture,
    client: MockerFixture,
) -> None:
    """
    Test the ``_build_where`` method.
    """
    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    columns: dict[str, Field] = {
        "is_food_order": Boolean(),
        "order_id": Integer(),
        "order_id__ordered_at__month": Date(),
    }
    adapter.grains["order_id__ordered_at__month"] = ("order_id__ordered_at", "month")

    assert adapter._build_where(columns, {}) == []

    with pytest.raises(ValueError):
        adapter._build_where(columns, {"is_food_order": Filter()})

    with pytest.raises(ImpossibleFilterError):
        adapter._build_where(columns, {"is_food_order": Impossible()})

    assert adapter._build_where(columns, {"is_food_order": Equal(True)}) == [
        {"sql": "{{ Dimension('is_food_order') }} = TRUE"},
    ]
    assert adapter._build_where(columns, {"is_food_order": NotEqual(True)}) == [
        {"sql": "{{ Dimension('is_food_order') }} != TRUE"},
    ]
    assert adapter._build_where(columns, {"is_food_order": IsNull()}) == [
        {"sql": "{{ Dimension('is_food_order') }} IS NULL"},
    ]
    assert adapter._build_where(columns, {"is_food_order": IsNotNull()}) == [
        {"sql": "{{ Dimension('is_food_order') }} IS NOT NULL"},
    ]
    assert adapter._build_where(columns, {"order_id": Range(10, 20)}) == [
        {
            "sql": "{{ Dimension('order_id') }} > 10 AND {{ Dimension('order_id') }} < 20",
        },
    ]
    assert adapter._build_where(
        columns,
        {"order_id": Range(10, 20, include_start=True, include_end=True)},
    ) == [
        {
            "sql": "{{ Dimension('order_id') }} >= 10 AND {{ Dimension('order_id') }} <= 20",
        },
    ]
    assert adapter._build_where(columns, {"order_id": Range(10, None)}) == [
        {
            "sql": "{{ Dimension('order_id') }} > 10",
        },
    ]
    assert adapter._build_where(columns, {"order_id": Range(None, 20)}) == [
        {
            "sql": "{{ Dimension('order_id') }} < 20",
        },
    ]
    assert adapter._build_where(
        columns,
        {"order_id__ordered_at__month": Range(None, datetime(2023, 1, 1, 12, 0))},
    ) == [
        {
            "sql": (
                "{{ TimeDimension('order_id__ordered_at', 'month') }}"
                " < '2023-01-01T12:00:00'"
            ),
        },
    ]


def test_timestamp() -> None:
    """
    Test ``Timestamp``.
    """
    assert Timestamp().parse(pd.Timestamp("2023-01-01 12:00")) == datetime(
        2023,
        1,
        1,
        12,
        0,
    )
    assert Timestamp().parse(None) is None
    assert Timestamp().format(datetime(2023, 1, 1, 12, 0)) == pd.Timestamp(
        "2023-01-01 12:00",
    )
    assert Timestamp().format(None) is None
    assert (
        Timestamp().quote(pd.Timestamp("2023-01-01 12:00")) == "'2023-01-01T12:00:00'"
    )
    assert Timestamp().quote(None) == "NULL"


def test_build_column_from_dimension(mocker: MockerFixture) -> None:
    """
    Test the ``_build_column_from_dimension`` method.
    """
    mocker.patch.object(DbtMetricFlowAPI, "_set_columns", return_value={})
    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    mocker.patch.object(
        adapter,
        "_run_query",
        side_effect=ProgrammingError("Something happened"),
    )
    assert adapter._build_column_from_dimension("dim") == Unknown(
        filters=[Equal, NotEqual, IsNull, IsNotNull],
        order=Order.ANY,
        exact=True,
    )

    mocker.patch.object(
        adapter,
        "_run_query",
        return_value=base64.b64encode(b"hello"),
    )
    ipc = mocker.patch("pyarrow.ipc")

    with ipc.open_stream() as reader:
        reader.schema = pa.schema([("dim", pa.bool_())])
        assert adapter._build_column_from_dimension("dim") == Boolean(
            filters=[Equal, NotEqual, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.int32())])
        assert adapter._build_column_from_dimension("dim") == Integer(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.float32())])
        assert adapter._build_column_from_dimension("dim") == DecimalField(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.string())])
        assert adapter._build_column_from_dimension("dim") == String(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.timestamp("ms"))])
        assert adapter._build_column_from_dimension("dim") == Timestamp(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.date32())])
        assert adapter._build_column_from_dimension("dim") == Date(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.time32("ms"))])
        assert adapter._build_column_from_dimension("dim") == Time(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.decimal128(10))])
        assert adapter._build_column_from_dimension("dim") == DecimalField(
            filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

        reader.schema = pa.schema([("dim", pa.binary())])
        assert adapter._build_column_from_dimension("dim") == Unknown(
            filters=[Equal, NotEqual, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )


def test_run_query_error(client_with_bad_request: MockerFixture) -> None:
    """
    Test errors happening in ``run_query``.
    """
    with pytest.raises(InternalError) as excinfo:
        DbtMetricFlowAPI(
            "https://semantic-layer.cloud.getdbt.com/api/graphql",
            "dbtc_XXX",
            123456,
        )
    assert (
        str(excinfo.value)
        == "Variable '$environmentId' of required type 'BigInt!' was not provided."
    )


def test_time_dimension_aliases(mocker: MockerFixture) -> None:
    """
    Test that time dimensions are created for each valid grain.
    """
    GraphqlClient = mocker.patch("shillelagh.adapters.api.dbt_metricflow.GraphqlClient")
    GraphqlClient().execute.return_value = {
        "data": {
            "metrics": [
                {
                    "description": "Sum of the product revenue for each "
                    "order item. Excludes tax.",
                    "dimensions": [
                        {
                            "expr": None,
                            "name": "metric_time",
                            "description": "The metric time",
                            "queryableGranularities": [
                                "DAY",
                                "WEEK",
                                "MONTH",
                                "QUARTER",
                                "YEAR",
                            ],
                            "type": "TIME",
                        },
                    ],
                    "name": "revenue",
                    "type": "SIMPLE",
                },
                {
                    "description": "Sum of cost for each order item.",
                    "dimensions": [
                        {
                            "expr": None,
                            "name": "metric_time",
                            "description": "The metric time",
                            "queryableGranularities": [
                                "DAY",
                                "WEEK",
                                "MONTH",
                                "QUARTER",
                                "YEAR",
                            ],
                            "type": "TIME",
                        },
                    ],
                    "name": "order_cost",
                    "type": "SIMPLE",
                },
            ],
        },
    }

    mocker.patch.object(
        DbtMetricFlowAPI,
        "_build_column_from_dimension",
        return_value=Unknown(),
    )

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    assert adapter.dimensions == {
        "metric_time__day": "The metric time",
        "metric_time__week": "The metric time",
        "metric_time__month": "The metric time",
        "metric_time__year": "The metric time",
        "metric_time__quarter": "The metric time",
    }
    assert adapter.grains == {
        "metric_time__day": ("metric_time", "day"),
        "metric_time__week": ("metric_time", "week"),
        "metric_time__month": ("metric_time", "month"),
        "metric_time__quarter": ("metric_time", "quarter"),
        "metric_time__year": ("metric_time", "year"),
    }


def test_build_groupbys(
    mocker: MockerFixture,
    client: MockerFixture,
) -> None:
    """
    Test the ``_build_groupbys`` method.
    """
    mocker.patch("time.sleep")

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    assert adapter.dimensions == {
        "order_id__is_food_order": "A boolean indicating if this order included any food items.",
    }
    assert adapter._build_groupbys({"order_id__is_food_order"}) == [
        {"name": "order_id__is_food_order"},
    ]


def test_build_orderbys(
    mocker: MockerFixture,
    client: MockerFixture,
) -> None:
    """
    Test the ``_build_orderbys`` method.
    """
    mocker.patch("time.sleep")

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    assert adapter.metrics == {"orders": "Count of orders."}
    assert adapter.dimensions == {
        "order_id__is_food_order": "A boolean indicating if this order included any food items.",
    }
    assert adapter._build_orderbys(
        [
            ("order_id__is_food_order", Order.ASCENDING),
            ("orders", Order.DESCENDING),
        ],
        [{"name": "order_id__is_food_order"}],
    ) == [
        {"descending": False, "groupBy": {"name": "order_id__is_food_order"}},
        {"descending": True, "metric": {"name": "orders"}},
    ]


def test_build_orderbys_error(
    mocker: MockerFixture,
    client: MockerFixture,
) -> None:
    """
    Test the ``_build_orderbys`` method when it encounters an error.
    """
    mocker.patch("time.sleep")

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    with pytest.raises(ProgrammingError) as excinfo:
        adapter._build_orderbys(
            [
                ("order_id__is_food_order", Order.ASCENDING),
                ("orders", Order.DESCENDING),
                ("invalid", Order.DESCENDING),
            ],
            [{"name": "order_id__is_food_order"}],
        )
    assert str(excinfo.value) == "Invalid order by column: invalid"


@pytest.mark.slow_integration_test
def test_integration(adapter_kwargs: dict[str, Any]) -> None:
    """
    Full integration test running a query.
    """
    connection = connect(":memory:", adapter_kwargs=adapter_kwargs)
    cursor = connection.cursor()

    sql = """
        SELECT
            order_id__ordered_at__month AS order_id__ordered_at,
            order_id__is_food_order AS order_id__is_food_order,
            orders as orders
        FROM "https://semantic-layer.cloud.getdbt.com/"
        WHERE
            order_id__ordered_at >= '2017-01-01T00:00:00.000000'
            AND order_id__ordered_at < '2018-01-01T00:00:00.000000'
            AND order_id__is_food_order = true
        GROUP BY
            order_id__ordered_at__month,
            order_id__is_food_order
        ORDER BY order_id__ordered_at DESC
        LIMIT 10000
    """
    cursor.execute(sql)
    data = cursor.fetchall()
    assert data == [
        (datetime(2017, 8, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("2347")),
        (datetime(2017, 7, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("2057")),
        (datetime(2017, 6, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("2378")),
        (datetime(2017, 5, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("2497")),
        (datetime(2017, 4, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("2101")),
        (datetime(2017, 3, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("1894")),
        (datetime(2017, 2, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("1127")),
        (datetime(2017, 1, 1, 0, 0, tzinfo=timezone.utc), True, Decimal("1185")),
    ]


def test_find_cursor(mocker: MockerFixture) -> None:
    """
    Test the ``find_cursor`` helper.
    """
    assert find_cursor() is None

    cursor = Cursor(mocker.MagicMock(), [], {})
    assert find_cursor() == cursor

    def nested() -> None:
        assert find_cursor() == cursor

    nested()


def test_get_metrics_for_dimensions(mocker: MockerFixture) -> None:
    """
    Test the ``_get_metrics_for_dimensions`` helper.
    """
    GraphqlClient: MagicMock = mocker.patch(
        "shillelagh.adapters.api.dbt_metricflow.GraphqlClient",
    )
    GraphqlClient().execute.side_effect = [
        *INIT_PAYLOAD,
        {
            "data": {
                "metricsForDimensions": [
                    {"name": "revenue"},
                    {"name": "order_cost"},
                    {"name": "median_revenue"},
                    {"name": "food_revenue"},
                    {"name": "food_revenue_pct"},
                    {"name": "revenue_growth_mom"},
                    {"name": "order_gross_profit"},
                    {"name": "cumulative_revenue"},
                    {"name": "order_total"},
                    {"name": "large_order"},
                    {"name": "orders"},
                    {"name": "food_orders"},
                    {"name": "customers_with_orders"},
                    {"name": "new_customer"},
                ],
            },
        },
    ]

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )
    assert adapter._get_metrics_for_dimensions({"order_id__is_food_order"}) == {
        "food_revenue_pct",
        "median_revenue",
        "food_orders",
        "food_revenue",
        "order_gross_profit",
        "cumulative_revenue",
        "large_order",
        "new_customer",
        "order_total",
        "order_cost",
        "revenue_growth_mom",
        "customers_with_orders",
        "revenue",
        "orders",
    }


def test_get_dimensions_for_metrics(mocker: MockerFixture) -> None:
    """
    Test the ``_get_dimensions_for_metrics`` helper.
    """
    GraphqlClient: MagicMock = mocker.patch(
        "shillelagh.adapters.api.dbt_metricflow.GraphqlClient",
    )
    GraphqlClient().execute.side_effect = [
        *INIT_PAYLOAD,
        {
            "data": {
                "dimensions": [
                    {"name": "customer__customer_name"},
                    {"name": "customer__customer_type"},
                    {"name": "customer__first_ordered_at"},
                    {"name": "customer__last_ordered_at"},
                    {"name": "location__location_name"},
                    {"name": "location__opened_at"},
                    {"name": "metric_time"},
                    {"name": "order_id__is_drink_order"},
                    {"name": "order_id__is_food_order"},
                    {"name": "order_id__order_total_dim"},
                    {"name": "order_id__ordered_at"},
                ],
            },
        },
    ]

    adapter = DbtMetricFlowAPI(
        "https://semantic-layer.cloud.getdbt.com/api/graphql",
        "dbtc_XXX",
        123456,
    )

    adapter.dimensions = {
        "order_id__is_food_order": "A boolean indicating if this order included any food items.",
        "metric_time__day": "The metric time",
        "metric_time__week": "The metric time",
        "metric_time__month": "The metric time",
        "metric_time__year": "The metric time",
        "metric_time__quarter": "The metric time",
    }
    adapter.grains = {
        "metric_time__day": ("metric_time", "day"),
        "metric_time__week": ("metric_time", "week"),
        "metric_time__month": ("metric_time", "month"),
        "metric_time__quarter": ("metric_time", "quarter"),
        "metric_time__year": ("metric_time", "year"),
    }

    assert adapter._get_dimensions_for_metrics({"orders"}) == {
        "order_id__is_food_order",
        "metric_time__day",
        "metric_time__week",
        "metric_time__month",
        "metric_time__year",
        "metric_time__quarter",
    }



"""
An adapter to WeatherAPI (https://www.weatherapi.com/).
"""

import logging
import urllib.parse
from collections.abc import Iterator
from datetime import date, datetime, timedelta, timezone
from typing import Any, Optional, Union, cast

import dateutil.parser
import dateutil.tz

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ImpossibleFilterError
from shillelagh.fields import DateTime, Float, IntBoolean, Integer, Order, String
from shillelagh.filters import Filter, Impossible, Operator, Range
from shillelagh.lib import get_session
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

INITIAL_COST = 0
FETCHING_COST = 1000


def combine_time_filters(bounds: dict[str, Filter]) -> Range:
    """
    Combine both time filters together.

    The adapter has two time columns that can be used to filter the data, "time" as
    a timestamp and "time_epoch" as a float. We convert the latter to a timestamp and
    combine the two filters into a single ``Range``.
    """
    time_range = bounds.get("time", Range())
    time_epoch_range = bounds.get("time_epoch", Range())

    if isinstance(time_range, Impossible) or isinstance(time_epoch_range, Impossible):
        raise ImpossibleFilterError()

    if not isinstance(time_range, Range) or not isinstance(time_epoch_range, Range):
        raise Exception("Invalid filter")  # pylint: disable=broad-exception-raised

    # convert time_epoch range to datetime so we can combine it
    # with the time range
    time_epoch_range.start = (
        datetime.fromtimestamp(time_epoch_range.start, tz=timezone.utc)
        if time_epoch_range.start is not None
        else None
    )
    time_epoch_range.end = (
        datetime.fromtimestamp(time_epoch_range.end, tz=timezone.utc)
        if time_epoch_range.end is not None
        else None
    )

    # combine time ranges together and check if the result is a valid range
    time_range += time_epoch_range
    if isinstance(time_range, Impossible):
        raise ImpossibleFilterError()

    return cast(Range, time_range)


class WeatherAPI(Adapter):
    """
    An adapter for WeatherAPI (https://www.weatherapi.com/).

    The adapter expects an URL like::

        https://api.weatherapi.com/v1/history.json?key=$key&q=$location

    Where ``$key`` is an API key (available for free), and ``$location`` is a
    freeform value that can be a US Zipcode, UK Postcode, Canada Postalcode,
    IP address, Latitude/Longitude (decimal degree) or city name.
    """

    safe = True

    # Since the adapter doesn't return exact data (see the time columns below)
    # implementing limit/offset is not worth the trouble.
    supports_limit = False
    supports_offset = False

    # These two columns can be used to filter the results from the API. We
    # define them as inexact since we will retrieve data for the whole day,
    # even if specific hours are requested. The post-filtering will be done
    # by the backend.
    time = DateTime(filters=[Range], order=Order.ASCENDING, exact=False)
    time_epoch = Float(filters=[Range], order=Order.ASCENDING, exact=False)

    temp_c = Float()
    temp_f = Float()
    is_day = IntBoolean()
    wind_mph = Float()
    wind_kph = Float()
    wind_degree = Integer()
    wind_dir = String()
    pressure_mb = Float()
    pressure_in = Float()
    precip_mm = Float()
    precip_in = Float()
    humidity = Integer()
    cloud = Integer()
    feelslike_c = Float()
    feelslike_f = Float()
    windchill_c = Float()
    windchill_f = Float()
    heatindex_c = Float()
    heatindex_f = Float()
    dewpoint_c = Float()
    dewpoint_f = Float()
    will_it_rain = IntBoolean()
    chance_of_rain = String()
    will_it_snow = IntBoolean()
    chance_of_snow = String()
    vis_km = Float()
    vis_miles = Float()
    gust_mph = Float()
    gust_kph = Float()

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        """https://api.weatherapi.com/v1/history.json?key=XXX&q=94158"""
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)
        return (
            parsed.netloc == "api.weatherapi.com"
            and parsed.path == "/v1/history.json"
            and "q" in query_string
            and ("key" in query_string or "api_key" in kwargs)
        )

    @staticmethod
    def parse_uri(uri: str) -> Union[tuple[str], tuple[str, str]]:
        parsed = urllib.parse.urlparse(uri)
        query_string = urllib.parse.parse_qs(parsed.query)
        location = query_string["q"][0]

        # key can be passed in the URL or via connection arguments
        if "key" in query_string:
            return (location, query_string["key"][0])
        return (location,)

    def __init__(self, location: str, api_key: str, window: int = 7):
        super().__init__()

        self.location = location
        self.api_key = api_key
        self.window = window

        # use a cache, since the adapter does a lot of similar API requests,
        # and the data should rarely (never?) change
        self._session = get_session(
            request_headers={},
            cache_name="weatherapi_cache",
            expire_after=timedelta(minutes=3),
        )

    def get_cost(
        self,
        filtered_columns: list[tuple[str, Operator]],
        order: list[tuple[str, RequestedOrder]],
    ) -> float:
        cost = INITIAL_COST

        # if the operator is ``Operator.EQ`` we only need to fetch 1 day of data;
        # otherwise we potentially need to fetch "window" days of data
        for _, operator in filtered_columns:
            weight = 1 if operator == Operator.EQ else self.window
            cost += FETCHING_COST * weight

        return cost

    def get_data(  # pylint: disable=too-many-locals
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        **kwargs: Any,
    ) -> Iterator[Row]:
        # combine filters from the two time columns
        try:
            time_range = combine_time_filters(bounds)
        except ImpossibleFilterError:
            return

        today = date.today()
        first = today - timedelta(days=self.window - 1)
        start = time_range.start.date() if time_range.start else first
        end = time_range.end.date() if time_range.end else today
        _logger.debug("Range is %s to %s", start, end)

        # download data from every today from [start, end]
        while start <= end:
            url = "https://api.weatherapi.com/v1/history.json"
            params = {"key": self.api_key, "q": self.location, "dt": start}

            query_string = urllib.parse.urlencode(params)
            _logger.info("GET %s?%s", url, query_string)

            response = self._session.get(url, params=params)
            if response.ok:
                payload = response.json()
                local_timezone = dateutil.tz.gettz(payload["location"]["tz_id"])
                for record in payload["forecast"]["forecastday"][0]["hour"]:
                    row = {column: record[column] for column in self.get_columns()}
                    row["time"] = dateutil.parser.parse(record["time"]).replace(
                        tzinfo=local_timezone,
                    )
                    row["rowid"] = int(row["time_epoch"])
                    _logger.debug(row)
                    yield row

            start += timedelta(days=1)



"""
Tests for shillelagh.fields.
"""

import datetime
import decimal
import sys
from typing import Union

import pytest

from shillelagh.adapters.registry import registry
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import (
    Blob,
    Boolean,
    Date,
    DateTime,
    Decimal,
    FastISODateTime,
    Field,
    Float,
    IntBoolean,
    Integer,
    ISODate,
    ISODateTime,
    ISOTime,
    Order,
    String,
    StringBlob,
    StringBoolean,
    StringDate,
    StringDateTime,
    StringDecimal,
    StringDuration,
    StringTime,
    Time,
    Unknown,
)
from shillelagh.filters import Equal
from shillelagh.types import BINARY, DATETIME, NUMBER, STRING

from .fakes import FakeAdapter


def test_comparison() -> None:
    """
    Test comparing fields.
    """
    field1 = String(filters=[Equal], order=Order.ASCENDING, exact=True)
    field2 = String(filters=[Equal], order=Order.ASCENDING, exact=True)
    field3 = String(filters=[Equal], order=Order.ASCENDING, exact=False)

    assert field1 == field2
    assert field1 != field3
    assert field1 != 42

    assert String(filters=[Equal], order=Order.ANY, exact=True) != Integer(
        filters=[Equal],
        order=Order.ANY,
        exact=True,
    )


def test_integer() -> None:
    """
    Test ``Integer``.
    """
    assert Integer().parse(1) == 1
    assert Integer().parse(None) is None
    assert Integer().format(1) == 1
    assert Integer().format(None) is None
    assert Integer().quote("1") == "1"  # type: ignore
    assert Integer().quote(None) == "NULL"


def test_float() -> None:
    """
    Test ``Float``.
    """
    assert Float().parse(1.0) == 1.0
    assert Float().parse(None) is None
    assert Float().format(1.0) == 1.0
    assert Float().format(None) is None
    assert Float().quote("1.0") == "1.0"  # type: ignore
    assert Float().quote(None) == "NULL"


def test_string() -> None:
    """
    Test ``String``.
    """
    assert String().parse("1.0") == "1.0"
    assert String().parse(None) is None
    assert String().format("test") == "test"
    assert String().format(None) is None
    assert String().quote("1.0") == "'1.0'"
    assert String().quote("O'Malley's") == "'O''Malley''s'"
    assert String().quote(None) == "NULL"


def test_date() -> None:
    """
    Test ``Date``.
    """
    assert Date().parse(datetime.date(2020, 1, 1)) == datetime.date(2020, 1, 1)
    assert Date().parse(None) is None
    assert Date().format(datetime.date(2020, 1, 1)) == datetime.date(2020, 1, 1)
    assert Date().format(None) is None
    assert Date().quote(datetime.date(2020, 1, 1)) == "'2020-01-01'"
    assert Date().quote(None) == "NULL"


def test_isodate() -> None:
    """
    Test ``ISODate``.
    """
    assert ISODate().parse("2020-01-01") == datetime.date(2020, 1, 1)
    assert ISODate().parse(None) is None
    assert ISODate().parse("invalid") is None
    assert ISODate().format(datetime.date(2020, 1, 1)) == "2020-01-01"
    assert ISODate().format(None) is None
    assert ISODate().quote("2020-01-01") == "'2020-01-01'"
    assert ISODate().quote(None) == "NULL"


def test_string_date() -> None:
    """
    Test ``StringDate``.
    """
    assert StringDate().parse("2020-01-01") == datetime.date(2020, 1, 1)
    assert StringDate().parse("2020-01-01T00:00+00:00") == datetime.date(
        2020,
        1,
        1,
    )
    assert StringDate().parse(None) is None
    assert StringDate().parse("invalid") is None


def test_time() -> None:
    """
    Test ``Time``.
    """
    assert Time().parse(
        datetime.time(12, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.time(12, 0, tzinfo=datetime.timezone.utc)
    assert Time().parse(None) is None
    assert Time().format(
        datetime.time(12, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.time(12, 0, tzinfo=datetime.timezone.utc)
    assert Time().format(None) is None
    assert (
        Time().quote(datetime.time(12, 0, tzinfo=datetime.timezone.utc))
        == "'12:00:00+00:00'"
    )
    assert Time().quote(None) == "NULL"


def test_isotime() -> None:
    """
    Test ``ISOTime``.
    """
    assert ISOTime().parse("12:00+00:00") == datetime.time(
        12,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert ISOTime().parse("12:00") == datetime.time(
        12,
        0,
    )
    assert ISOTime().parse(None) is None
    assert ISOTime().parse("invalid") is None
    assert (
        ISOTime().format(datetime.time(12, 0, tzinfo=datetime.timezone.utc))
        == "12:00:00+00:00"
    )
    assert ISOTime().format(None) is None
    assert ISOTime().quote("12:00:00+00:00") == "'12:00:00+00:00'"
    assert ISOTime().quote(None) == "NULL"


def test_string_time() -> None:
    """
    Test ``StringTime``.
    """
    assert StringTime().parse("12:00+00:00") == datetime.time(
        12,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert StringTime().parse("12:00") == datetime.time(
        12,
        0,
    )
    assert StringTime().parse(None) is None
    assert StringTime().parse("invalid") is None


def test_datetime() -> None:
    """
    Test ``DateTime``.
    """
    assert DateTime().parse(
        datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)
    assert DateTime().parse(None) is None
    assert DateTime().format(
        datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
    ) == datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc)
    assert DateTime().format(None) is None
    assert (
        DateTime().quote(
            datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
        )
        == "'2020-01-01T12:00:00+00:00'"
    )
    assert DateTime().quote(None) == "NULL"


def test_isodatetime() -> None:
    """
    Test ``ISODateTime``.
    """
    assert ISODateTime().parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert ISODateTime().parse("2020-01-01T12:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
    )
    assert ISODateTime().parse(None) is None
    assert ISODateTime().parse("invalid") is None
    assert (
        ISODateTime().format(
            datetime.datetime(2020, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc),
        )
        == "2020-01-01T12:00:00+00:00"
    )
    assert ISODateTime().format(None) is None
    assert (
        ISODateTime().quote("2020-01-01T12:00:00+00:00")
        == "'2020-01-01T12:00:00+00:00'"
    )
    assert ISODateTime().quote(None) == "NULL"


def test_string_datetime() -> None:
    """
    Test ``StringDateTime``.
    """
    assert StringDateTime().parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert StringDateTime().parse("2020-01-01T12:00Z") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert StringDateTime().parse("2020-01-01T12:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
    )
    assert StringDateTime().parse(None) is None
    assert StringDateTime().parse("invalid") is None


def test_boolean() -> None:
    """
    Test ``Boolean``.
    """
    assert Boolean().parse(True) is True
    assert Boolean().parse(False) is False
    assert Boolean().parse(None) is None
    assert Boolean().format(True) is True
    assert Boolean().format(False) is False
    assert Boolean().format(None) is None
    assert Boolean().quote(True) == "TRUE"
    assert Boolean().quote(False) == "FALSE"
    assert Boolean().quote(None) == "NULL"


def test_int_boolean() -> None:
    """
    Test ``IntBoolean``.
    """
    assert IntBoolean().parse(1) is True
    assert IntBoolean().parse(0) is False
    assert IntBoolean().parse(10) is True
    assert IntBoolean().parse(None) is None
    assert IntBoolean().format(True) == 1
    assert IntBoolean().format(False) == 0
    assert IntBoolean().format(None) is None
    assert IntBoolean().quote(1) == "1"
    assert IntBoolean().quote(0) == "0"
    assert IntBoolean().quote(None) == "NULL"


def test_string_boolean() -> None:
    """
    Test ``StringBoolean``.
    """
    assert StringBoolean().parse("TRUE") is True
    assert StringBoolean().parse("FALSE") is False
    with pytest.raises(ValueError) as expected_err:
        StringBoolean().parse("B")
    assert str(expected_err.value) == "invalid truth value b"
    assert StringBoolean().parse(None) is None
    assert StringBoolean().format(True) == "TRUE"
    assert StringBoolean().format(False) == "FALSE"
    assert StringBoolean().format(None) is None
    assert StringBoolean().quote("TRUE") == "TRUE"
    assert StringBoolean().quote("FALSE") == "FALSE"
    assert StringBoolean().quote(None) == "NULL"


def test_blob() -> None:
    """
    Test ``Blob``.
    """
    assert Blob().parse(b"test") == b"test"
    assert Blob().parse(None) is None
    assert Blob().format(b"test") == b"test"
    assert Blob().format(None) is None
    assert Blob().quote(b"test") == "X'74657374'"
    assert Blob().quote(None) == "NULL"


def test_string_blob() -> None:
    """
    Test ``StringBlob``.
    """
    assert StringBlob().parse("74657374") == b"test"
    assert StringBlob().parse(None) is None
    assert StringBlob().format(b"test") == "74657374"
    assert StringBlob().format(None) is None
    assert StringBlob().quote("74657374") == "X'74657374'"
    assert StringBlob().quote(None) == "NULL"


def test_type_code() -> None:
    """
    Test typecodes for Python DB API.
    """
    assert Integer == NUMBER
    assert Float == NUMBER
    assert String == STRING
    assert Date == DATETIME
    assert Time == DATETIME
    assert DateTime == DATETIME
    assert ISODate == DATETIME
    assert ISOTime == DATETIME
    assert ISODateTime == DATETIME
    assert Blob == BINARY
    assert Boolean == NUMBER

    assert NUMBER != 1


def test_duration() -> None:
    """
    Test ``Duration``.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = "CREATE TABLE test_table (a DURATION)"
    cursor.execute(sql)

    sql = "INSERT INTO test_table (a) VALUES (?)"
    cursor.execute(sql, (datetime.timedelta(hours=1),))

    sql = "SELECT * FROM test_table"
    cursor.execute(sql)
    assert cursor.fetchall() == [(datetime.timedelta(hours=1),)]


def test_string_duration() -> None:
    """
    Test ``StringDuration``.
    """
    assert StringDuration().parse("12:34:56") == datetime.timedelta(
        hours=12,
        minutes=34,
        seconds=56,
    )
    assert StringDuration().parse("12:34:56.789012") == datetime.timedelta(
        hours=12,
        minutes=34,
        seconds=56,
        microseconds=789012,
    )
    assert StringDuration().parse(None) is None
    assert StringDuration().parse("2 days, 4:00:00") == datetime.timedelta(
        days=2,
        hours=4,
    )
    assert (
        StringDuration().format(datetime.timedelta(hours=12, minutes=34, seconds=56))
        == "12:34:56"
    )
    assert (
        StringDuration().format(
            datetime.timedelta(hours=12, minutes=34, seconds=56, microseconds=789012),
        )
        == "12:34:56.789012"
    )
    assert StringDuration().format(None) is None
    assert (
        StringDuration().format(datetime.timedelta(days=2, hours=4))
        == "2 days, 4:00:00"
    )
    assert (
        StringDuration().quote(
            datetime.timedelta(  # type: ignore
                hours=12,
                minutes=34,
                seconds=56,
                microseconds=789012,
            ),
        )
        == "'12:34:56.789012'"
    )
    assert StringDuration().quote(None) == "NULL"

    connection = connect(":memory:")
    cursor = connection.cursor()

    sql = "CREATE TABLE test_table (a DURATION)"
    cursor.execute(sql)

    sql = "INSERT INTO test_table (a) VALUES (?)"
    cursor.execute(sql, (datetime.timedelta(hours=1),))

    sql = "SELECT * FROM test_table"
    cursor.execute(sql)
    assert cursor.fetchall() == [(datetime.timedelta(hours=1),)]


def test_polymorphic_field() -> None:
    """
    Test for a polymorphic field.
    """

    class IntegerOrString(Field[Union[int, str], Union[int, str]]):
        """
        A column that can be an integer or a string.
        """

        type = "TEXT"
        db_api_type = "STRING"

    class CustomFakeAdapter(FakeAdapter):
        """
        A simple adapter with an ``IntegerOrString`` column.
        """

        secret = IntegerOrString()

        def __init__(self):
            super().__init__()

            self.data = [
                {"rowid": 0, "name": "Alice", "age": 20, "pets": 0, "secret": 42},
                {"rowid": 1, "name": "Bob", "age": 23, "pets": 3, "secret": "XXX"},
            ]

    registry.add("dummy", CustomFakeAdapter)

    connection = connect(":memory:", ["dummy"], isolation_level="IMMEDIATE")
    cursor = connection.cursor()

    cursor.execute('SELECT * FROM "dummy://"')
    assert cursor.fetchall() == [(20, "Alice", 0, 42), (23, "Bob", 3, "XXX")]


def test_fastisodatetime() -> None:
    """
    Test ``FastISODateTime``.
    """
    assert FastISODateTime().parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert FastISODateTime().parse(None) is None
    assert FastISODateTime().parse("2020-01-01T12:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
    )

    if sys.version_info < (3, 11):
        with pytest.raises(ProgrammingError) as excinfo:
            FastISODateTime().parse("2020-01-01T12:00Z")
        assert str(excinfo.value) == 'Unable to parse "2020-01-01T12:00Z"'

    with pytest.raises(ProgrammingError) as excinfo:
        FastISODateTime().parse("invalid")
    assert str(excinfo.value) == 'Unable to parse "invalid"'


def test_stringdecimal() -> None:
    """
    Test ``StringDecimal``.
    """
    assert StringDecimal().parse("1.23") == decimal.Decimal("1.23")
    assert StringDecimal().parse(None) is None
    assert StringDecimal().format(decimal.Decimal("1.23")) == "1.23"
    assert StringDecimal().format(None) is None
    assert StringDecimal().quote("1.23") == "1.23"
    assert StringDecimal().quote(None) == "NULL"


def test_decimal() -> None:
    """
    Test ``Decimal``.
    """
    assert Decimal().parse(decimal.Decimal("1.23")) == decimal.Decimal("1.23")
    assert Decimal().parse(None) is None
    assert Decimal().format(decimal.Decimal("1.23")) == decimal.Decimal("1.23")
    assert Decimal().format(None) is None
    assert Decimal().quote(decimal.Decimal("1.23")) == "1.23"
    assert Decimal().quote(None) == "NULL"


def test_unkown() -> None:
    """
    Test ``Unknown``.
    """
    assert Unknown().parse(1) == 1
    assert Unknown().parse("1") == "1"
    assert Unknown().parse(True) is True
    assert Unknown().parse(None) is None

    assert Unknown().format(1) == 1
    assert Unknown().format("1") == "1"
    assert Unknown().format(True) is True
    assert Unknown().format(None) is None

    assert Unknown().quote(1) == "1"
    assert Unknown().quote("1") == "'1'"
    assert Unknown().quote(True) == "1"
    assert Unknown().quote(None) == "NULL"



"""
A simple example showing the Socrata adapter.
"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(":memory:")
    cursor = connection.cursor()

    SQL = """
    SELECT date, administered_dose1_recip_4
    FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
    WHERE location = 'US'
    ORDER BY date DESC
    LIMIT 10
    """
    for row in cursor.execute(SQL):
        print(row)



"""
Tests for the {{ cookiecutter.adapter_name|replace(' ', '') }} adapter.
"""
import requests

from shillelagh.backends.apsw.db import connect


def test_{{ cookiecutter.slug }}(mocker, requests_mock):
    """
    Run SQL against the adapter.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()

    # Mock network requests.
    adapter = requests_mock.Adapter()
    adapter.register_uri("GET", "https://api.example.com/", json={"hello": "world"})

    # And mock the session object.
    session = requests.Session()
    session.mount("https://api.example.com/", adapter)
    mocker.patch(
        "shillelagh.adapters.{{ cookiecutter.adapter_type|lower }}.{{ cookiecutter.slug }}.{{ cookiecutter.adapter_name|replace(' ', '') }}{{ cookiecutter.adapter_type }}._get_session",
        return_value=session,
    )

    # Replace this with a URI supported by the adapter.
    sql = """
        SELECT * FROM "https://api.example.com/"
    """
    cursor.execute(sql)
    assert cursor.fetchall() == []



"""
A simple example querying the dbt MetricFlow.
"""

from shillelagh.backends.apsw.db import connect

if __name__ == "__main__":
    connection = connect(
        ":memory:",
        adapter_kwargs={
            "dbtmetricflowapi": {
                "service_token": "dbtc_XXX",
                "environment_id": 123456,
            },
        },
    )
    cursor = connection.cursor()

    SQL = """
    SELECT orders, order_id__is_food_order
    FROM "https://semantic-layer.cloud.getdbt.com/"
    LIMIT 10
    """
    for row in cursor.execute(SQL):
        print(row)



"""
An adapter for querying dbt Metric Flow metrics.

This is meant to be used with Apache Superset and has only been tested with the SQL that
it produces, eg:

    SELECT
        order_id__ordered_at__month AS order_id__ordered_at,
        order_id__is_food_order AS order_id__is_food_order,
        orders as orders
    FROM "https://semantic-layer.cloud.getdbt.com/"
    WHERE
        order_id__ordered_at >= '2017-01-01T00:00:00.000000'
        AND order_id__ordered_at < '2018-01-01T00:00:00.000000'
        AND order_id__is_food_order = true
    GROUP BY
        order_id__ordered_at__month,
        order_id__is_food_order
    ORDER BY order_id__ordered_at DESC
    LIMIT 10000

"""

import base64
import datetime
import inspect
import logging
import re
import time
from collections import defaultdict
from collections.abc import Iterator
from typing import Any, Optional, TypedDict, cast
from urllib.parse import urlparse

import pandas as pd
import pyarrow as pa
import sqlglot
from python_graphql_client import GraphqlClient
from sqlglot import expressions as exp
from sqlglot.optimizer.qualify_columns import qualify_columns
from sqlglot.optimizer.scope import traverse_scope

from shillelagh.adapters.base import Adapter
from shillelagh.backends.apsw.db import Cursor
from shillelagh.exceptions import ImpossibleFilterError, InternalError, ProgrammingError
from shillelagh.fields import (
    Boolean,
    Date,
    Decimal,
    Field,
    Filter,
    Integer,
    Order,
    String,
    Time,
    Unknown,
)
from shillelagh.filters import Equal, Impossible, IsNotNull, IsNull, NotEqual, Range
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)

CUSTOM_URL_PATTERN = re.compile(
    r"""
        (?P<id>
            [a-zA-Z0-9]
            (?:
                [a-zA-Z0-9-]{0,61}
                [a-zA-Z0-9]
            )?
        )
        \.
        (?P<region>
            [a-zA-Z0-9]
            (?:
                [a-zA-Z0-9-]{0,61}
                [a-zA-Z0-9]
            )?
        )
        \.dbt.com
        $
    """,
    re.VERBOSE,
)

LIST_METRICS = """
    query GetMetrics($environmentId: BigInt!) {
        metrics(environmentId: $environmentId) {
            name
            description
            type
            dimensions {
                name
                description
                queryableGranularities
                type
                expr
            }
        }
    }
"""

CREATE_QUERY = """
    mutation GetData(
        $environmentId: BigInt!,
        $metrics: [MetricInput!]!,
        $where: [WhereInput!]!,
        $groupBy: [GroupByInput!]!,
        $orderBy: [OrderByInput!]!,
        $limit: Int,
    ) {
        createQuery(
            environmentId: $environmentId,
            metrics: $metrics,
            where: $where,
            groupBy: $groupBy,
            orderBy: $orderBy,
            limit: $limit,
        ) {
            queryId
        }
    }
"""

POLL_RESULTS = """
    query PollResults(
        $environmentId: BigInt!,
        $queryId: String!,
    ) {
        query(
            environmentId: $environmentId,
            queryId: $queryId,
        ) {
            sql
            status
            error
            arrowResult
        }
    }
"""

METRICS_FOR_DIMENSIONS = """
    query GetDimensions(
        $environmentId: BigInt!,
        $dimensions: [GroupByInput!]!,
    ) {
        metricsForDimensions(
            environmentId: $environmentId,
            dimensions: $dimensions,
        ) {
            name
        }
    }
"""

DIMENSIONS_FOR_METRICS = """
    query GetMetrics(
        $environmentId: BigInt!,
        $metrics: [MetricInput!]!,
    ) {
        dimensions(
            environmentId: $environmentId,
            metrics: $metrics,
        ) {
            name
        }
    }
"""


def find_cursor() -> Optional[Cursor]:
    """
    Inspects the stack up to find the cursor that called the adapter.

    This is ugly. It makes me want to quit tech and become a gardener.
    """
    for frame in inspect.stack():
        for value in frame.frame.f_locals.values():
            if isinstance(value, Cursor):
                return value

    return None


def extract_columns_from_sql(table: str, columns: set[str], sql: str) -> set[str]:
    """
    Parse the SQL and extract requested columns.
    """
    schema = {table: {column: "unused" for column in columns}}
    ast = sqlglot.parse_one(sql)
    ast = qualify_columns(ast, schema=schema)
    return {
        column.name
        for scope in traverse_scope(ast)
        for column in scope.columns
        if isinstance(scope.sources.get(column.table), exp.Table)
        and column.table == table
    }


class Timestamp(Field[pd.Timestamp, datetime.datetime]):
    """
    Pandas Timestamp.
    """

    type = "TIMESTAMP"
    db_api_type = "DATETIME"

    def parse(self, value: Optional[pd.Timestamp]) -> Optional[datetime.datetime]:
        return value.to_pydatetime() if value is not None else None

    def format(self, value: Optional[datetime.datetime]) -> Optional[pd.Timestamp]:
        return pd.Timestamp(value) if value is not None else None

    def quote(self, value: Optional[pd.Timestamp]) -> str:
        if value is None:
            return "NULL"
        return f"'{value.to_pydatetime().isoformat()}'"


class WhereInput(TypedDict):
    """
    Where parameter in the GraphQL API.
    """

    sql: str


class GroupByInput(TypedDict, total=False):
    """
    Group by parameter in the GraphQL API.
    """

    name: str
    grain: Optional[str]  # DAY, WEEK, MONTH, QUARTER, YEAR
    datePart: Optional[str]  # YEAR, QUARTER, MONTH, DAY, DOW, DOY


class MetricInput(TypedDict):
    """
    Metric input parameter in the GraphQL API.
    """

    name: str


class OrderByInput(TypedDict, total=False):
    """
    Order by parameter in the GraphQL API.
    """

    metric: MetricInput
    groupBy: GroupByInput
    descending: bool


def stream_to_dataframe(byte_string: str) -> pd.DataFrame:
    """
    Convert an Arrow stream to a Pandas DataFrame.
    """
    with pa.ipc.open_stream(base64.b64decode(byte_string)) as reader:
        return pa.Table.from_batches(reader, reader.schema).to_pandas()


class DbtMetricFlowAPI(Adapter):
    """
    An adapter for querying dbt Metric Flow metrics.
    """

    safe = True
    supports_limit = True
    supports_offset = False

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        """
        Is this a semantic layer URL?

        The adapter supports both the old single URL or custom user URLs.
        """
        parsed = urlparse(uri)
        return uri == "https://semantic-layer.cloud.getdbt.com/" or (
            parsed.scheme == "https" and bool(CUSTOM_URL_PATTERN.match(parsed.netloc))
        )

    @staticmethod
    def parse_uri(uri: str) -> tuple[str]:
        return (uri,)

    def __init__(self, table: str, service_token: str, environment_id: int):
        super().__init__()

        self.table = table
        self.environment_id = environment_id

        endpoint = self._get_endpoint(table)
        self.client = GraphqlClient(
            endpoint=endpoint,
            headers={"Authorization": f"Bearer {service_token}"},
        )

        self._set_columns()

    @staticmethod
    def _get_endpoint(url: str) -> str:
        """
        Return the GraphQL endpoint.
        """
        parsed = urlparse(url)
        if match := CUSTOM_URL_PATTERN.match(parsed.netloc):
            return f"https://{match['id']}.semantic-layer.{match['region']}.dbt.com/api/graphql"

        return "https://semantic-layer.cloud.getdbt.com/api/graphql"

    def _set_columns(self) -> None:
        payload = self.client.execute(
            query=LIST_METRICS,
            variables={"environmentId": self.environment_id},
        )

        self.columns: dict[str, Field] = {}
        self.metrics: dict[str, str] = {}
        self.dimensions: dict[str, str] = {}
        self.grains: dict[str, tuple[str, str]] = {}

        built_dimensions: set[str] = set()
        for metric in payload["data"]["metrics"]:
            self.columns[metric["name"]] = Decimal()
            self.metrics[metric["name"]] = metric["description"]

            for dimension in metric["dimensions"]:
                name = dimension["name"]
                if name in built_dimensions:
                    continue

                column = self._build_column_from_dimension(name)
                built_dimensions.add(name)

                # for time dimensions we create a dimension for each grain
                if dimension["type"] == "TIME":
                    for grain in dimension["queryableGranularities"]:
                        alias = f"{name}__{grain.lower()}"
                        self.columns[alias] = column
                        self.dimensions[alias] = dimension["description"]
                        self.grains[alias] = (name, grain.lower())
                else:
                    self.columns[name] = column
                    self.dimensions[name] = dimension["description"]

        self.columns = dict(sorted(self.columns.items()))

    def _run_query(self, **variables: Any) -> str:
        """
        Run a query, wait for it, and return Arrow payload.
        """
        payload = self.client.execute(query=CREATE_QUERY, variables=variables)
        if errors := payload.get("errors"):
            raise InternalError("\n\n".join(error["message"] for error in errors))

        query_id = payload["data"]["createQuery"]["queryId"]

        while True:
            payload = self.client.execute(
                query=POLL_RESULTS,
                variables={
                    "environmentId": self.environment_id,
                    "queryId": query_id,
                },
            )
            if error := payload["data"]["query"].get("error"):
                _logger.debug(payload["data"]["query"]["sql"])
                raise ProgrammingError(error)

            if payload["data"]["query"]["status"] == "SUCCESSFUL":
                break

            time.sleep(1)

        return cast(str, payload["data"]["query"]["arrowResult"])

    def _get_metrics_for_dimensions(self, dimensions: set[str]) -> set[str]:
        """
        Get metrics for a set of dimensions.
        """
        payload = self.client.execute(
            query=METRICS_FOR_DIMENSIONS,
            variables={
                "environmentId": self.environment_id,
                "dimensions": [{"name": dimension} for dimension in dimensions],
            },
        )

        return {metric["name"] for metric in payload["data"]["metricsForDimensions"]}

    def _get_dimensions_for_metrics(self, metrics: set[str]) -> set[str]:
        """
        Get dimensions for a set of metrics.
        """
        payload = self.client.execute(
            query=DIMENSIONS_FOR_METRICS,
            variables={
                "environmentId": self.environment_id,
                "metrics": [{"name": metric} for metric in metrics],
            },
        )

        reverse_grain: dict[str, set[str]] = defaultdict(set)
        for alias, (
            name,
            grain,  # pylint: disable=unused-variable
        ) in self.grains.items():
            reverse_grain[name].add(alias)

        dimensions: set[str] = set()
        for dimension in payload["data"]["dimensions"]:
            if dimension["name"] in self.dimensions:
                dimensions.add(dimension["name"])
            elif dimension["name"] in reverse_grain:
                dimensions.update(reverse_grain[dimension["name"]])

        return dimensions

    def _build_column_from_dimension(  # pylint: disable=too-many-return-statements
        self,
        name: str,
    ) -> Field:
        """
        Build a field from a dbt dimension.

        Unfortunately the API does not provide the type of the dimension, so we
        need to fetch each one and read the type from the Arrow response.
        """
        try:
            byte_string = self._run_query(
                environmentId=self.environment_id,
                metrics=[],
                where=[],
                groupBy=[{"name": name}],
                orderBy=[],
                limit=0,
            )
        except ProgrammingError:
            return Unknown(
                filters=[Equal, NotEqual, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )

        with pa.ipc.open_stream(base64.b64decode(byte_string)) as reader:
            field = reader.schema[0]

        if pa.types.is_boolean(field.type):
            return Boolean(
                filters=[Equal, NotEqual, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_integer(field.type):
            return Integer(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_floating(field.type):
            return Decimal(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_string(field.type):
            return String(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_timestamp(field.type):
            return Timestamp(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_date(field.type):
            return Date(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_time(field.type):
            return Time(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )
        if pa.types.is_decimal(field.type):
            return Decimal(
                filters=[Equal, NotEqual, Range, IsNull, IsNotNull],
                order=Order.ANY,
                exact=True,
            )

        return Unknown(
            filters=[Equal, NotEqual, IsNull, IsNotNull],
            order=Order.ANY,
            exact=True,
        )

    def _build_where(
        self,
        columns: dict[str, Field],
        bounds: dict[str, Filter],
    ) -> list[WhereInput]:
        """
        Build a ``WhereInput`` list from the bounds to filter in GraphQL.
        """
        where: list[WhereInput] = []
        for column_name, filter_ in bounds.items():
            if isinstance(filter_, Impossible):
                raise ImpossibleFilterError()

            if column_name in self.grains:
                base_column_name, grain = self.grains[column_name]
                ref = f"{{{{ TimeDimension('{base_column_name}', '{grain}') }}}}"
            else:
                ref = f"{{{{ Dimension('{column_name}') }}}}"

            field = columns[column_name]
            if isinstance(filter_, Equal):
                sql = f"{ref} = {field.quote(filter_.value)}"
            elif isinstance(filter_, NotEqual):
                sql = f"{ref} != {field.quote(filter_.value)}"
            elif isinstance(filter_, IsNull):
                sql = f"{ref} IS NULL"
            elif isinstance(filter_, IsNotNull):
                sql = f"{ref} IS NOT NULL"
            elif isinstance(filter_, Range):
                conditions = []
                if filter_.start is not None:
                    operator_ = ">=" if filter_.include_start else ">"
                    conditions.append(f"{ref} {operator_} {field.quote(filter_.start)}")
                if filter_.end is not None:
                    operator_ = "<=" if filter_.include_end else "<"
                    conditions.append(f"{ref} {operator_} {field.quote(filter_.end)}")
                sql = " AND ".join(conditions)
            else:
                raise ValueError(f"Invalid filter: {filter_}")

            where.append({"sql": sql})

        return where

    def _build_groupbys(
        self,
        requested_columns: set[str],
    ) -> list[GroupByInput]:
        """
        Build group bys based on the requested columns and the SQL query.
        """
        return [
            {"name": column}
            for column in requested_columns
            if column in self.dimensions
        ]

    def _build_orderbys(
        self,
        order: list[tuple[str, RequestedOrder]],
        groupbys: list[GroupByInput],
    ) -> list[OrderByInput]:
        """
        Build order bys based on the requested columns and the SQL query.
        """
        groupbys_map = {groupby["name"]: groupby for groupby in groupbys}

        orderbys = []
        for column, requested_order in order:
            orderby: OrderByInput = {
                "descending": requested_order == Order.DESCENDING,
            }

            if column in self.metrics:
                orderby["metric"] = {"name": column}
            elif column in self.dimensions:
                orderby["groupBy"] = groupbys_map[column]
            else:
                raise ProgrammingError(f"Invalid order by column: {column}")

            orderbys.append(orderby)

        return orderbys

    def get_columns(self) -> dict[str, Field]:
        return self.columns

    def get_data(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        limit: Optional[int] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        cursor = find_cursor()
        if cursor is None or cursor.operation is None:
            raise InternalError("Unable to get reference to cursor")

        # When the virtual table has more than 63 columns apsw will return any requested
        # columns in the 1-63 range, plus ALL columns from 64-, since the mask is a 64-bit
        # value. This doesn't work for dbt, since only related columns should be requested
        # together. To solve this we need to parse the SQL and extract the columns
        # ourselves.
        requested_columns = extract_columns_from_sql(
            self.table,
            set(self.columns),
            cursor.operation,
        )

        groupbys = self._build_groupbys(requested_columns)
        orderbys = self._build_orderbys(order, groupbys)

        byte_string = self._run_query(
            environmentId=self.environment_id,
            metrics=[
                {"name": column}
                for column in requested_columns
                if column in self.metrics
            ],
            where=self._build_where(self.columns, bounds),
            groupBy=groupbys,
            orderBy=orderbys,
            limit=limit,
        )

        # pylint: disable=invalid-name
        df = stream_to_dataframe(byte_string)
        df.rename_axis("rowid", inplace=True)

        yield from df.reset_index().to_dict(orient="records")



# pylint: disable=invalid-name, c-extension-no-member, unused-import
"""
A DB API 2.0 wrapper.
"""

import logging
import re
from typing import Any, Optional, Union, cast
from uuid import uuid4

import psycopg2
from psycopg2 import extensions

from shillelagh.adapters.base import Adapter
from shillelagh.adapters.registry import registry
from shillelagh.exceptions import (  # nopycln: import; pylint: disable=redefined-builtin
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    OperationalError,
    ProgrammingError,
    Warning,
)
from shillelagh.lib import (
    combine_args_kwargs,
    escape_identifier,
    find_adapter,
    serialize,
)
from shillelagh.types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)

__all__ = [
    "DatabaseError",
    "DataError",
    "Error",
    "IntegrityError",
    "InterfaceError",
    "InternalError",
    "OperationalError",
    "BINARY",
    "DATETIME",
    "NUMBER",
    "ROWID",
    "STRING",
    "Binary",
    "Date",
    "DateFromTicks",
    "Time",
    "TimeFromTicks",
    "Timestamp",
    "TimestampFromTicks",
    "Warning",
]

apilevel = "2.0"
threadsafety = 2
paramstyle = "pyformat"

NO_SUCH_TABLE = re.compile('relation "(.*?)" does not exist')
DEFAULT_SCHEMA = "main"

_logger = logging.getLogger(__name__)


class Cursor(extensions.cursor):  # pylint: disable=too-few-public-methods
    """
    A cursor that registers FDWs.
    """

    def __init__(
        self,
        *args: Any,
        adapters: dict[str, type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        schema: str,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self._adapters = list(adapters.values())
        self._adapter_map = {v: k for k, v in adapters.items()}
        self._adapter_kwargs = adapter_kwargs
        self.schema = schema

    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> Union["Cursor", extensions.cursor]:
        """
        Execute a query, automatically registering FDWs if necessary.
        """
        # which cursor should be returned
        cursor: Union["Cursor", extensions.cursor] = self

        while True:
            savepoint = uuid4()
            super().execute(f'SAVEPOINT "{savepoint}"')

            try:
                cursor = cast(extensions.cursor, super().execute(operation, parameters))
                break
            except psycopg2.errors.UndefinedTable as ex:  # pylint: disable=no-member
                message = ex.args[0]
                match = NO_SUCH_TABLE.match(message)
                if not match:
                    raise ProgrammingError(message) from ex

                # Postgres truncates the table name in the error message, so we need to
                # find it in the original query
                fragment = match.group(1)
                uri = self._get_table_uri(fragment, operation)
                if not uri:
                    raise ProgrammingError("Could not determine table name") from ex

                super().execute(f'ROLLBACK TO SAVEPOINT "{savepoint}"')
                self._create_table(uri)

        if uri := self._drop_table_uri(operation):
            adapter, args, kwargs = find_adapter(
                uri,
                self._adapter_kwargs,
                self._adapters,
            )
            instance = adapter(*args, **kwargs)
            instance.drop_table()

        return cursor

    def _get_table_uri(self, fragment: str, operation: str) -> Optional[str]:
        """
        Extract the table name from a query.
        """
        schema = re.escape(self.schema)
        fragment = re.escape(fragment)
        regexp = re.compile(
            rf'\b(FROM|INTO)\s+({schema}\.)?(?P<uri>"{fragment}.*?")',
            re.IGNORECASE,
        )
        if match := regexp.search(operation):
            return match.groupdict()["uri"].strip('"')

        return None

    def _drop_table_uri(self, operation: str) -> Optional[str]:
        """
        Build a ``DROP TABLE`` regexp.
        """
        schema = re.escape(self.schema)
        regexp = re.compile(
            r"^\s*DROP\s+TABLE\s+(IF\s+EXISTS\s+)?"
            rf'({schema}\.)?(?P<uri>(.*?)|(".*?"))\s*;?\s*$',
            re.IGNORECASE,
        )
        if match := regexp.match(operation):
            return match.groupdict()["uri"].strip('"')

        return None

    def _create_table(self, uri: str) -> None:
        """
        Register a FDW.
        """
        adapter, args, kwargs = find_adapter(uri, self._adapter_kwargs, self._adapters)
        formatted_args = serialize(combine_args_kwargs(adapter, *args, **kwargs))
        entrypoint = self._adapter_map[adapter]

        table_name = escape_identifier(uri)

        columns = adapter(*args, **kwargs).get_columns()
        if not columns:
            raise ProgrammingError(f"Virtual table {table_name} has no columns")

        quoted_columns = {k.replace('"', '""'): v for k, v in columns.items()}
        formatted_columns = ", ".join(
            f'"{k}" {v.type}' for (k, v) in quoted_columns.items()
        )

        super().execute(
            """
CREATE SERVER shillelagh foreign data wrapper multicorn options (
    wrapper 'shillelagh.backends.multicorn.fdw.MulticornForeignDataWrapper'
);
    """,
        )
        super().execute(
            f"""
CREATE FOREIGN TABLE "{table_name}" (
    {formatted_columns}
) server shillelagh options (
    adapter '{entrypoint}',
    args '{formatted_args}'
);
        """,
        )


class CursorFactory:  # pylint: disable=too-few-public-methods
    """
    Custom cursor factory.

    This returns a custom cursor that will auto register FDWs for the user.
    """

    def __init__(
        self,
        adapters: dict[str, type[Adapter]],
        adapter_kwargs: dict[str, dict[str, Any]],
        schema: str,
    ):
        self.schema = schema
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs

    def __call__(self, *args, **kwargs) -> Cursor:
        """
        Create a new cursor.
        """
        return Cursor(
            *args,
            adapters=self._adapters,
            adapter_kwargs=self._adapter_kwargs,
            schema=self.schema,
            **kwargs,
        )


def connect(  # pylint: disable=too-many-arguments
    dsn: Optional[str] = None,
    adapters: Optional[list[str]] = None,
    adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
    schema: str = DEFAULT_SCHEMA,
    **psycopg2_connection_kwargs: Any,
) -> extensions.connection:
    """
    Constructor for creating a connection to the database.

    Only safe adapters can be loaded. If no adapters are specified, all safe adapters are
    loaded.
    """
    adapter_kwargs = adapter_kwargs or {}
    enabled_adapters = {
        name: adapter
        for name, adapter in registry.load_all(adapters, safe=False).items()
        if adapter.safe
    }

    # replace entry point names with class names
    mapping = {
        name: adapter.__name__.lower() for name, adapter in enabled_adapters.items()
    }
    adapter_kwargs = {mapping[k]: v for k, v in adapter_kwargs.items() if k in mapping}

    cursor_factory = CursorFactory(
        enabled_adapters,
        adapter_kwargs,
        schema,
    )
    return psycopg2.connect(
        dsn,
        cursor_factory=cursor_factory,
        **psycopg2_connection_kwargs,
    )



"""
Simple multicorn2 test.

Multicorn2 is an extension for PostgreSQL that allows you to create foreign data wrappers
in Python. To use it, you need to install on the machine running Postgres the extension,
the multicorn2 package (not on (PyPI), and the shillelagh package.

If you want to play with it Shillelagh has a `docker-compose.yml` file that will run
Postgres with the extension and the Python packages. Just run:

    $ docker compose -f postgres/docker-compose.yml up

Then you can run this script.
"""

from sqlalchemy import create_engine, text

# the backend uses psycopg2 under the hood, so any valid connection string for it will
# work; just replace the scheme with `shillelagh+multicorn2`
engine = create_engine(
    "shillelagh+multicorn2://shillelagh:shillelagh123@localhost:5432/shillelagh",
)
connection = engine.connect()

SQL = text(
    'SELECT * FROM "https://docs.google.com/spreadsheets/d/'
    '1LcWZMsdCl92g7nA-D6qGRqg1T5TiHyuKJUY1u9XAnsk/edit#gid=0"',
)
for row in connection.execute(SQL):
    print(row)



"""
Tests for shillelagh.types.
"""

from datetime import date, datetime, time, timezone

from shillelagh.backends.apsw.db import connect
from shillelagh.fields import String
from shillelagh.types import (
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)


def test_types() -> None:
    """
    Test that native Python types can be used in queries.
    """
    connection = connect(":memory:")
    cursor = connection.cursor()

    cursor.execute(
        """
        CREATE TABLE test_types (
            type_date DATE,
            type_time TIME,
            type_timestamp TIMESTAMP,
            type_binary BLOB
        )
    """,
    )
    cursor.execute(
        "INSERT INTO test_types VALUES (?, ?, ?, ?)",
        (
            Date(2020, 1, 1),
            Time(0, 0, 0),
            Timestamp(2020, 1, 1, 0, 0, 0),
            Binary("🦥"),
        ),
    )
    cursor.execute("SELECT * FROM test_types")
    row = cursor.fetchone()
    assert row == (
        date(2020, 1, 1),
        time(0, 0, tzinfo=timezone.utc),
        datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
        b"\xf0\x9f\xa6\xa5",
    )

    cursor.execute(
        "INSERT INTO test_types VALUES (?, ?, ?, ?)",
        (
            DateFromTicks(1),
            TimeFromTicks(2),
            TimestampFromTicks(3),
            Binary("🦥"),
        ),
    )
    cursor.execute("SELECT * FROM test_types")
    rows = cursor.fetchall()
    assert rows == [
        (
            date(2020, 1, 1),
            time(0, 0, tzinfo=timezone.utc),
            datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
            b"\xf0\x9f\xa6\xa5",
        ),
        (
            date(1970, 1, 1),
            time(0, 0, 2, tzinfo=timezone.utc),
            datetime(1970, 1, 1, 0, 0, 3, tzinfo=timezone.utc),
            b"\xf0\x9f\xa6\xa5",
        ),
    ]


def test_comparison() -> None:
    """
    Test type comparison.
    """
    assert STRING == String
    assert (STRING == 1) is False



"""
Tests for the multicorn dialect.
"""

from pytest_mock import MockerFixture
from sqlalchemy.engine.url import make_url

from shillelagh.backends.multicorn import db
from shillelagh.backends.multicorn.db import Cursor
from shillelagh.backends.multicorn.dialects.base import (
    Multicorn2Dialect,
    get_adapter_for_table_name,
)
from shillelagh.exceptions import ProgrammingError

from ....fakes import FakeAdapter


def test_dbapi() -> None:
    """
    Test the ``dbapi`` and ``import_dbapi`` methods.
    """
    assert Multicorn2Dialect.dbapi() == Multicorn2Dialect.import_dbapi() == db


def test_create_connect_args() -> None:
    """
    Test ``create_connect_args``.
    """
    dialect = Multicorn2Dialect(["dummy"], {})
    assert dialect.create_connect_args(
        make_url(
            "shillelagh+multicorn2://shillelagh:shillelagh123@localhost:12345/shillelagh",
        ),
    ) == (
        [],
        {
            "adapter_kwargs": {},
            "adapters": ["dummy"],
            "user": "shillelagh",
            "password": "shillelagh123",
            "host": "localhost",
            "port": 12345,
            "dbname": "shillelagh",
        },
    )


def test_has_table(mocker: MockerFixture) -> None:
    """
    Test ``has_table``.
    """
    super = mocker.patch(  # pylint: disable=redefined-builtin
        "shillelagh.backends.multicorn.dialects.base.super",
        create=True,
    )
    has_table = mocker.MagicMock(name="has_table", return_value=False)
    super.return_value.has_table = has_table
    mocker.patch(
        "shillelagh.backends.multicorn.dialects.base.get_adapter_for_table_name",
        side_effect=[True, ProgrammingError('No adapter for table "dummy://".')],
    )
    connection = mocker.MagicMock()

    dialect = Multicorn2Dialect(["dummy"], {})
    assert dialect.has_table(connection, "dummy://") is True
    assert dialect.has_table(connection, "my_table") is False


def test_get_adapter_for_table_name(mocker: MockerFixture) -> None:
    """
    Test the ``get_adapter_for_table_name`` function.
    """
    mocker.patch("shillelagh.backends.multicorn.db.super", create=True)
    connection = mocker.MagicMock()
    connection.engine.raw_connection().cursor.return_value = Cursor(
        adapters={"dummy": FakeAdapter},
        adapter_kwargs={},
        schema="main",
    )

    assert isinstance(get_adapter_for_table_name(connection, "dummy://"), FakeAdapter)
